import os
import sys
from datetime import datetime,timedelta
import logging
import pathlib
import tempfile
import subprocess
import shutil
from typing import Union
from time import time
import glob

import numpy as np
import scipy as sp
from numba import jit, prange
import netCDF4 as nc
from netCDF4 import Dataset
import pandas as pd
import xarray as xr
from matplotlib.transforms import Bbox
import seawater as sw

from pyschism.mesh.base import Nodes, Elements
from pyschism.mesh.vgrid import Vgrid
from pyschism.forcing.hycom.hycom2schism import Nudge

logger = logging.getLogger(__name__)

def get_idxs(date, ds, bbox):

    time1=ds['time']
    times=nc.num2date(time1,units=time1.units,only_use_cftime_datetimes=False)
    
    lon=ds['lon'][:]
    lat=ds['lat'][:]
    lat_idxs=np.where((lat>=bbox.ymin-2.0)&(lat<=bbox.ymax+2.0))[0]
    lon_idxs=np.where((lon>=bbox.xmin-2.0) & (lon<=bbox.xmax+2.0))[0]
    lon=lon[lon_idxs]
    lat=lat[lat_idxs]
    lon_idx1=lon_idxs[0].item()
    lon_idx2=lon_idxs[-1].item()
    lat_idx1=lat_idxs[0].item()
    lat_idx2=lat_idxs[-1].item()
    
    for ilon in np.arange(len(lon)):
        if lon[ilon] > 180:
            lon[ilon] = lon[ilon]-360.
    x2, y2=transform_ll_to_cpp(lon, lat)

    idxs=np.where( date == times)[0]
    #check if time_idx is empty
    if len(idxs) == 0:
        #If there is missing data, use the data from the next days, the maximum searching days is 3. Otherwise, stop.
        for i in np.arange(0,3):
            date_before=(date + timedelta(days=int(i)+1)) #.astype(datetime)
            logger.info(f'Try replacing the missing data from {date_before}')
            idxs=np.where(date_before == times)[0]
            if len(idxs) == 0:
                continue
            else:
                break
    if len(idxs) ==0:
        logger.info(f'No date for date {date}')
        sys.exit()
    time_idx=idxs.item()  
    logger.info(f'time_idx is {time_idx}')

    return time_idx, lon_idx1, lon_idx2, lat_idx1, lat_idx2, x2, y2

def get_idxs2(date, ds):

    times = pd.to_datetime(ds['time'])
    #times=nc.num2date(time1,units=time1.units,only_use_cftime_datetimes=False)
    idxs=np.where( date == times)[0]
    time_idx=idxs.item()  

    #lon=ds['lon'][:]
    #lat=ds['lat'][:]
    lon=ds.lon.values
    lat=ds.lat.values
    for ilon in np.arange(len(lon)):
        if lon[ilon] > 180:
            lon[ilon] = lon[ilon]-360.
    x2, y2=transform_ll_to_cpp(lon, lat)

    return time_idx, x2, y2

def transform_ll_to_cpp(lon, lat, lonc=-77.07, latc=24.0):
    longitude=lon/180*np.pi
    latitude=lat/180*np.pi
    radius=6378206.4
    loncc=lonc/180*np.pi
    latcc=latc/180*np.pi
    lon_new=[radius*(longitude[i]-loncc)*np.cos(latcc) for i in np.arange(len(longitude))]
    lat_new=[radius*latitude[i] for i in np.arange(len(latitude))]

    return np.array(lon_new), np.array(lat_new)

def interp_to_points_3d(dep, y2, x2, bxyz, val):
    idxs = np.where(abs(val) > 10000)
    val[idxs] = float('nan')

    val_fd = sp.interpolate.RegularGridInterpolator((dep,y2,x2),np.squeeze(val),'linear', bounds_error=False, fill_value = float('nan'))
    val_int = val_fd(bxyz)
    idxs = np.isnan(val_int)
    if np.sum(idxs) != 0:
        val_int[idxs] = sp.interpolate.griddata(bxyz[~idxs,:], val_int[~idxs], bxyz[idxs,:],'nearest')
    idxs = np.isnan(val_int)
    if np.sum(idxs) != 0:
        logger.info(f'There is still missing value for {val}')
        sys.exit()
    return val_int

def interp_to_points_2d(y2, x2, bxy, val):
    idxs = np.where(abs(val) > 10000)
    val[idxs] = float('nan')

    val_fd = sp.interpolate.RegularGridInterpolator((y2,x2),np.squeeze(val),'linear', bounds_error=False, fill_value = float('nan'))
    val_int = val_fd(bxy)
    idxs = np.isnan(val_int)
    if np.sum(idxs) != 0:
        val_int[idxs] = sp.interpolate.griddata(bxy[~idxs,:], val_int[~idxs], bxy[idxs,:],'nearest')
    idxs = np.isnan(val_int)
    if np.sum(idxs) != 0:
        logger.info(f'There is still missing value for {val}')
        sys.exit()
    return val_int

class OpenBoundaryInventory:

    def __init__(self, hgrid, vgrid=None):
        self.hgrid = hgrid
        self.vgrid = Vgrid.default() if vgrid is None else vgrid

    def fetch_data(self, outdir: Union[str, os.PathLike], start_date, rnday, elev2D=True, TS=True, UV=True, adjust2D=False, lats=None, msl_shifts=None, cached=False): 
        outdir = pathlib.Path(outdir)

        self.start_date = start_date
        self.rnday=rnday
        self.timevector=np.arange(
            self.start_date,
            self.start_date + timedelta(days=self.rnday+1),
            timedelta(days=1)).astype(datetime)

        #Get open boundary 
        gdf=self.hgrid.boundaries.open.copy()
        opbd=[]
        for boundary in gdf.itertuples():
            opbd.extend(list(boundary.indexes))
        blon = self.hgrid.coords[opbd,0]
        blat = self.hgrid.coords[opbd,1]
        NOP = len(blon)

        #calculate zcor for 3D
        if TS or UV:
            vd=Vgrid.open(self.vgrid)
            sigma=vd.sigma

            #get bathymetry
            depth = self.hgrid.values

            #compute zcor
            zcor = depth[:,None]*sigma
            nvrt=zcor.shape[1]

        #create netcdf
        ntimes=self.rnday+1
        nComp1=1
        nComp2=2
        one=1
        ndt=np.zeros([ntimes])

        if elev2D:
            timeseries_el=np.zeros([ntimes,NOP,nComp1])
            #create netcdf 
            dst_elev = Dataset(outdir / 'elev2D.th.nc', 'w', format='NETCDF4')
            #dimensions
            dst_elev.createDimension('nOpenBndNodes', NOP)
            dst_elev.createDimension('one', one)
            dst_elev.createDimension('time', None)
            dst_elev.createDimension('nLevels', one)
            dst_elev.createDimension('nComponents', nComp1)

            #variables
            dst_elev.createVariable('time_step', 'f', ('one',))
            dst_elev['time_step'][:] = 86400

            dst_elev.createVariable('time', 'f', ('time',))
            #dst_elev['time'][:] = ndt

            dst_elev.createVariable('time_series', 'f', ('time', 'nOpenBndNodes', 'nLevels', 'nComponents'))
            #dst_elev['time_series'][:,:,:,:] = timeseries_el

        if TS:
            #timeseries_s=np.zeros([ntimes,NOP,nvrt,nComp1])
            dst_salt = Dataset(outdir / 'SAL_3D.th.nc', 'w', format='NETCDF4')
            #dimensions
            dst_salt.createDimension('nOpenBndNodes', NOP)
            dst_salt.createDimension('one', one)
            dst_salt.createDimension('time', None)
            dst_salt.createDimension('nLevels', nvrt)
            dst_salt.createDimension('nComponents', nComp1)
            #variables
            dst_salt.createVariable('time_step', 'f', ('one',))
            dst_salt['time_step'][:] = 86400

            dst_salt.createVariable('time', 'f', ('time',))
            #dst_salt['time'][:] = ndt

            dst_salt.createVariable('time_series', 'f', ('time', 'nOpenBndNodes', 'nLevels', 'nComponents'))

            #temp
            timeseries_t=np.zeros([ntimes,NOP,nvrt,nComp1])

            dst_temp =  Dataset(outdir / 'TEM_3D.th.nc', 'w', format='NETCDF4')
            #dimensions
            dst_temp.createDimension('nOpenBndNodes', NOP)
            dst_temp.createDimension('one', one)
            dst_temp.createDimension('time', None)
            dst_temp.createDimension('nLevels', nvrt)
            dst_temp.createDimension('nComponents', nComp1)
            #variables
            dst_temp.createVariable('time_step', 'f', ('one',))
            dst_temp['time_step'][:] = 86400

            dst_temp.createVariable('time', 'f', ('time',))
            #dst_temp['time'][:] = ndt

            dst_temp.createVariable('time_series', 'f', ('time', 'nOpenBndNodes', 'nLevels', 'nComponents'))
            #dst_temp['time_series'][:,:,:,:] = timeseries_t

        if UV:
            #timeseries_uv=np.zeros([ntimes,NOP,nvrt,nComp2])
            dst_uv = Dataset(outdir / 'uv3D.th.nc', 'w', format='NETCDF4')
            #dimensions
            dst_uv.createDimension('nOpenBndNodes', NOP)
            dst_uv.createDimension('one', one)
            dst_uv.createDimension('time', None)
            dst_uv.createDimension('nLevels', nvrt)
            dst_uv.createDimension('nComponents', nComp2)
            #variables
            dst_uv.createVariable('time_step', 'f', ('one',))
            dst_uv['time_step'][:] = 86400

            dst_uv.createVariable('time', 'f', ('time',))
            #dst_uv['time'][:] = ndt
            
            dst_uv.createVariable('time_series', 'f', ('time', 'nOpenBndNodes', 'nLevels', 'nComponents'))
            #dst_uv['time_series'][:,:,:,:] = timeseries_uv

        if cached:
            logger.info('**** Use cached data*****')
            files = glob.glob('rtofs_*.nc')
            files.sort()
            ncfiles = files[-9:]
            logger.info(f'files are {ncfiles}')

        else:
            logger.info('**** Accessing RTOFS data*****')
            baseurl = f'http://nomads.ncep.noaa.gov:80/dods/rtofs/rtofs_global'

        t0=time()
        for it, date in enumerate(self.timevector):
            logger.info(f'Fetching data for {date}')

            #loop over each open boundary
            ind1 = 0
            ind2 = 0
            for boundary in gdf.itertuples():

                opbd = list(boundary.indexes)
                ind1 = ind2
                ind2 = ind1 + len(opbd)
                blon = self.hgrid.coords[opbd,0]
                blat = self.hgrid.coords[opbd,1]
                xi,yi = transform_ll_to_cpp(blon, blat)
                bxy = np.c_[yi, xi]

                if TS or UV:
                    zcor2=zcor[opbd,:]
                    idxs=np.where(zcor2 > 5500)
                    zcor2[idxs]=5500.0-1.0e-6

                    #construct schism grid
                    x2i=np.tile(xi,[nvrt,1]).T
                    y2i=np.tile(yi,[nvrt,1]).T
                    bxyz=np.c_[zcor2.reshape(np.size(zcor2)),y2i.reshape(np.size(y2i)),x2i.reshape(np.size(x2i))]

                xmin, xmax = np.min(blon), np.max(blon)
                ymin, ymax = np.min(blat), np.max(blat)

                # convert hgrid lon [180, 180) to [0, 360)
                xmin = xmin + 360. if xmin < 0 else xmin
                xmax = xmax + 360. if xmax < 0 else xmax
                bbox = Bbox.from_extents(xmin, ymin, xmax, ymax)
                logger.info(f'xmin is {xmin}, xmax is {xmax}')

                logger.info('****Interpolation starts****')

                #ndt[it]=it*24*3600.

                if elev2D:
                    #ssh
                    if cached:
                        ds = xr.open_mfdataset(ncfiles, decode_timedelta=True)
                        time_idx, x2, y2 = get_idxs2(date, ds)
                        logger.info(f'ssh: time_idx for {date} is {time_idx}')
                        ssh = ds.ssh.values[time_idx, :, :]

                    else:
                        ssh_url = f'{baseurl}{self.start_date.strftime("%Y%m%d")}/rtofs_glo_2ds_forecast_3hrly_diag'
                        ds = Dataset(ssh_url)
                        logger.info(f'ssh_url is {ssh_url}')
                        time_idx, lon_idx1, lon_idx2, lat_idx1, lat_idx2, x2, y2 = get_idxs(date, ds, bbox)
                        ssh = np.squeeze(ds['ssh'][time_idx+1,0,lat_idx1:lat_idx2+1,lon_idx1:lon_idx2+1])

                    ssh_int = interp_to_points_2d(y2, x2, bxy, ssh)
                    dst_elev['time'][it] = it*24*3600.
                    if adjust2D:
                        elev_adjust = np.interp(blat, lats, msl_shifts)
                        dst_elev['time_series'][it,ind1:ind2,0,0] = ssh_int + elev_adjust
                    else:
                        dst_elev['time_series'][it,ind1:ind2,0,0] = ssh_int 
                    ds.close()

                if TS:
                    #salt
                    if cached:
                        ds = xr.open_mfdataset(ncfiles, decode_timedelta=True)
                        time_idx, x2, y2 = get_idxs2(date, ds)
                        logger.info(f'salinity: time_idx for {date} is {time_idx}')
                        salt = ds.salinity.values[time_idx,:,:,:]
                    else:
                        salt_url = f'{baseurl}{self.start_date.strftime("%Y%m%d")}/rtofs_glo_3dz_forecast_daily_salt'
                        ds = Dataset(salt_url)
                        logger.info(f'salt_url is {salt_url}')
                        time_idx, lon_idx1, lon_idx2, lat_idx1, lat_idx2, x2, y2 = get_idxs(date, ds, bbox)
                        salt = np.squeeze(ds['salinity'][time_idx+1,:,lat_idx1:lat_idx2+1,lon_idx1:lon_idx2+1])
                    dep = ds['lev'][:]

                    salt_int = interp_to_points_3d(dep, y2, x2, bxyz, salt)
                    salt_int = salt_int.reshape(zcor2.shape)
                    #timeseries_s[it,:,:,0]=salt_int
                    dst_salt['time'][it] = it*24*3600.
                    dst_salt['time_series'][it,ind1:ind2,:,0] = salt_int

                    ds.close()

                    #temp
                    if cached:
                        ds = xr.open_mfdataset(ncfiles, decode_timedelta=True)
                        time_idx, x2, y2 = get_idxs2(date, ds)
                        logger.info(f'temperature: time_idx for {date} is {time_idx}')
                        temp = ds.temperature.values[time_idx,:,:,:]
                    else:
                        temp_url = f'{baseurl}{self.start_date.strftime("%Y%m%d")}/rtofs_glo_3dz_forecast_daily_temp'
                        ds = Dataset(temp_url)
                        logger.info(f'temp_url is {temp_url}')
                        temp = np.squeeze(ds['temperature'][time_idx+1,:,lat_idx1:lat_idx2+1,lon_idx1:lon_idx2+1])
                    dep = ds['lev'][:]

                    temp_int = interp_to_points_3d(dep, y2, x2, bxyz, temp)
                    temp_int = temp_int.reshape(zcor2.shape)
                    #timeseries_t[it,:,:,0]=temp_int
                    dst_temp['time'][it] = it*24*3600.
                    dst_temp['time_series'][it,ind1:ind2,:,0] = temp_int

                    ds.close()

                if UV:
                    if cached:
                        ds = xr.open_mfdataset(ncfiles, decode_timedelta=True)
                        time_idx, x2, y2 = get_idxs2(date, ds)
                        logger.info(f'uvel: time_idx for {date} is {time_idx}')
                        uvel = ds.u.values[time_idx,:,:,:]
                    else:
                        uvel_url = f'{baseurl}{self.start_date.strftime("%Y%m%d")}/rtofs_glo_3dz_forecast_daily_uvel'
                        ds = Dataset(uvel_url)
                        logger.info(f'uvel_url is {uvel_url}')
                        time_idx, lon_idx1, lon_idx2, lat_idx1, lat_idx2, x2, y2 = get_idxs(date, ds, bbox)
                        uvel = np.squeeze(ds['u'][time_idx+1,:,lat_idx1:lat_idx2+1,lon_idx1:lon_idx2+1])
                    dep = ds['lev'][:]

                    dst_uv['time'][it] = it*24*3600.
                    #uvel
                    uvel_int = interp_to_points_3d(dep, y2, x2, bxyz, uvel)
                    uvel_int = uvel_int.reshape(zcor2.shape)
                    dst_uv['time_series'][it,ind1:ind2,:,0] = uvel_int
                    ds.close()

                    #vvel
                    if cached:
                        ds = xr.open_mfdataset(ncfiles, decode_timedelta=True)
                        time_idx, x2, y2 = get_idxs2(date, ds)
                        logger.info(f'vvel: time_idx for {date} is {time_idx}')
                        vvel = ds.v.values[time_idx,:,:,:]
                    else:
                        vvel_url = f'{baseurl}{self.start_date.strftime("%Y%m%d")}/rtofs_glo_3dz_forecast_daily_vvel'
                        ds = Dataset(vvel_url)
                        logger.info(f'vvel_url is {vvel_url}')
                        vvel = np.squeeze(ds['v'][time_idx+1,:,lat_idx1:lat_idx2+1,lon_idx1:lon_idx2+1])
                    dep = ds['lev'][:]

                    vvel_int = interp_to_points_3d(dep, y2, x2, bxyz, vvel)
                    vvel_int = vvel_int.reshape(zcor2.shape)
                    dst_uv['time_series'][it,ind1:ind2,:,1] = vvel_int
                    #timeseries_uv[it,:,:,1]=vvel_int

        logger.info(f'Writing *th.nc takes {time()-t0} seconds')

class NudgeTS:

    def __init__(self):
        pass

    def fetch_data(self, outdir: Union[str, os.PathLike], hgrid, vgrid, start_date, rnday, include, cached=False):

        outdir = pathlib.Path(outdir)

        timevector = np.arange(start_date, start_date + timedelta(days=rnday+1), \
            timedelta(days=1)).astype(datetime)

        vd = Vgrid.open(vgrid)
        sigma = vd.sigma

        #get coords of SCHISM
        loni=hgrid.nodes.coords[:,0]
        lati=hgrid.nodes.coords[:,1]

        #get bathymetry
        depth = hgrid.values

        #compute zcor
        zcor = depth[:,None]*sigma
        nvrt=zcor.shape[1]

        #Get open nudge array 
        nlon = hgrid.coords[include, 0]
        nlat = hgrid.coords[include, 1]
        xi,yi = transform_ll_to_cpp(nlon, nlat)
        bxy = np.c_[yi, xi]

        zcor2=zcor[include,:]
        idxs=np.where(zcor2 > 5500)
        zcor2[idxs]=5500.0-1.0e-6

        #construct schism grid
        x2i=np.tile(xi,[nvrt,1]).T
        y2i=np.tile(yi,[nvrt,1]).T
        bxyz=np.c_[zcor2.reshape(np.size(zcor2)),y2i.reshape(np.size(y2i)),x2i.reshape(np.size(x2i))]
        logger.info('Computing SCHISM zcor is done!')

        #allocate output variables
        nNode=len(include)
        one=1
        ntimes=rnday+1

        timeseries_s=np.zeros([ntimes,nNode,nvrt,one])
        timeseries_t=np.zeros([ntimes,nNode,nvrt,one])
        ndt=np.zeros([ntimes])


        xmin, xmax = np.min(nlon), np.max(nlon)
        ymin, ymax = np.min(nlat), np.max(nlat)

        # convert hgrid lon [180, 180) to [0, 360)
        xmin = xmin + 360. if xmin < 0 else xmin
        xmax = xmax + 360. if xmax < 0 else xmax
        bbox = Bbox.from_extents(xmin, ymin, xmax, ymax)
        logger.info(f'xmin is {xmin}, xmax is {xmax}')

        if cached:
            logger.info('**** Use cached data*****')
            files = glob.glob('rtofs_*.nc')
            files.sort()
            ncfiles = files[-9:]
            logger.info(f'files are {ncfiles}')
        else:
            logger.info('**** Accessing RTOFS data*****')
            baseurl = f'http://nomads.ncep.noaa.gov:80/dods/rtofs/rtofs_global'

        t0=time()
        for it, date in enumerate(timevector):

            logger.info(f'Fetching data for {date}')

            ndt[it]=it
             
            #salt
            if cached:
                ds = xr.open_mfdataset(ncfiles, decode_timedelta=True)
                time_idx, x2, y2 = get_idxs2(date, ds)
                salt = ds.salinity.values[time_idx,:,:,:]
            else:
                salt_url = f'{baseurl}{start_date.strftime("%Y%m%d")}/rtofs_glo_3dz_forecast_daily_salt'
                ds = Dataset(salt_url)
                time_idx, lon_idx1, lon_idx2, lat_idx1, lat_idx2, x2, y2 = get_idxs(date, ds, bbox)
                #It seems that salt=0 at time_idx=0 in rtofs 
                salt = np.squeeze(ds['salinity'][time_idx+1,:,lat_idx1:lat_idx2+1,lon_idx1:lon_idx2+1])
            dep = ds['lev'][:]

            salt_int = interp_to_points_3d(dep, y2, x2, bxyz, salt)
            salt_int = salt_int.reshape(zcor2.shape)
            #timeseries_s[it,:,:,0]=salt_int
            timeseries_s[it,:,:,0] = salt_int

            ds.close()

            #temp
            if cached:
                ds = xr.open_mfdataset(ncfiles, decode_timedelta=True)
                time_idx, x2, y2 = get_idxs2(date, ds)
                temp = ds.temperature.values[time_idx,:,:,:]
            else:
                temp_url = f'{baseurl}{start_date.strftime("%Y%m%d")}/rtofs_glo_3dz_forecast_daily_temp'
                ds = Dataset(temp_url)
                temp = np.squeeze(ds['temperature'][time_idx+1,:,lat_idx1:lat_idx2+1,lon_idx1:lon_idx2+1])
            dep = ds['lev'][:]

            temp_int = interp_to_points_3d(dep, y2, x2, bxyz, temp)
            temp_int = temp_int.reshape(zcor2.shape)
            #timeseries_t[it,:,:,0]=temp_int
            timeseries_t[it,:,:,0] = temp_int
            ds.close()

        with Dataset(outdir / 'SAL_nu.nc', 'w', format='NETCDF4') as dst:
        #dimensions
            dst.createDimension('node', nNode)
            dst.createDimension('nLevels', nvrt)
            dst.createDimension('one', one)
            dst.createDimension('time', None)
        #variables
            dst.createVariable('time', 'f', ('time',))
            dst['time'][:] = ndt

            dst.createVariable('map_to_global_node', 'i4', ('node',))
            dst['map_to_global_node'][:] = include+1

            dst.createVariable('tracer_concentration', 'f', ('time', 'node', 'nLevels', 'one'))
            dst['tracer_concentration'][:,:,:,:] = timeseries_s

        with Dataset(outdir / 'TEM_nu.nc', 'w', format='NETCDF4') as dst:
        #dimensions
            dst.createDimension('node', nNode)
            dst.createDimension('nLevels', nvrt)
            dst.createDimension('one', one)
            dst.createDimension('time', None)
        #variables
            dst.createVariable('time', 'f', ('time',))
            dst['time'][:] = ndt

            dst.createVariable('map_to_global_node', 'i4', ('node',))
            dst['map_to_global_node'][:] = include+1

            dst.createVariable('tracer_concentration', 'f', ('time', 'node', 'nLevels', 'one'))
            dst['tracer_concentration'][:,:,:,:] = timeseries_t

        logger.info(f'Writing *_nu.nc takes {time()-t0} seconds')

class DownloadRTOFS:

    def __init__(self):
        pass

    def fetch_data(self, startdate, rnday=9):
        logger.info(f'startdate is {startdate}')
        idx_x1 = 2687
        idx_x2 = 2714
        idx_y1 = 1181
        idx_y2 = 1634
        logger.info(f'idx_x1 is {idx_x1}, idx_x2 is {idx_x2}, idx_y1 is {idx_y1}, idx_y2 is {idx_y2}')
        enddate = startdate + timedelta(days=rnday)
        datevector = np.arange(startdate, enddate, np.timedelta64(1, 'D'),dtype='datetime64')
        logger.info(datevector)
        datevector = pd.to_datetime(datevector)

        url_2d = f'http://nomads.ncep.noaa.gov:80/dods/rtofs/rtofs_global{startdate.strftime("%Y%m%d")}/rtofs_glo_2ds_forecast_3hrly_diag'
        url_3d = f'http://nomads.ncep.noaa.gov:80/dods/rtofs/rtofs_global{startdate.strftime("%Y%m%d")}/rtofs_glo_3dz_forecast_daily_'

        for it, date in enumerate(datevector):
            if it == 0:
                continue

            salt = []
            temp = []
            uvel = []
            vvel = []
            ssh = []
            vars = {'salt': ['salinity', salt], 'temp': ['temperature', temp], 'uvel': ['u', uvel], 'vvel': ['v', vvel],'ssh': ['ssh', ssh]}

            foutname = f'rtofs_{date.strftime("%Y%m%d")}.nc'
            logger.info(f'filename is {foutname}')

            attrs = {}
            for ivar, (key, var) in enumerate(vars.items()):
                if key == 'ssh':
                    url = url_2d
                else:
                    url = url_3d + key
                logger.info(f'url is {url}')
                ds = xr.open_dataset(url)
                if ivar == 0:
                    time=ds['time'][it].values
                    lon = ds['lon'][idx_x1:idx_x2+1].astype('float32')
                    lat = ds['lat'][idx_y1:idx_y2+1].astype('float32')
                    depth = ds['lev'][:].astype('float32')
                values = ds[var[0]]

                if key == 'ssh':
                    var[1].append(values[it*8, 0, idx_y1:idx_y2+1, idx_x1:idx_x2+1].astype('float32'))
                else: 
                    var[1].append(values[it, :, idx_y1:idx_y2+1, idx_x1:idx_x2+1].astype('float32'))
                attrs[key] = ds[var[0]].attrs
                #print(attrs)

            fout = xr.Dataset({
                    'salinity': (['time','lev', 'lat', 'lon'], np.array(salt)),
                    'temperature': (['time','lev', 'lat', 'lon'], np.array(temp)),
                    'u': (['time','lev', 'lat', 'lon'], np.array(uvel)),
                    'v': (['time','lev', 'lat', 'lon'], np.array(vvel)),
                    'ssh': (['time','lat', 'lon'], np.array(ssh)),
                    },
                    coords = {
                        'time': np.atleast_1d(time),
                        'lev': depth,
                        'lat': lat,
                        'lon': lon,
                    })
            fout.salinity.attrs = attrs['salt']
            fout.temperature.attrs = attrs['temp']
            fout.u.attrs = attrs['uvel']
            fout.v.attrs = attrs['vvel']
            fout.ssh.attrs = attrs['ssh']
            fout.time.encoding['units'] = 'days since 0001-01-01 00:00:00'
            
            fout.to_netcdf(foutname, 'w', unlimited_dims='time')



from abc import ABC
from datetime import datetime, timedelta
from functools import lru_cache
import logging
import os
import pathlib
from typing import Union


import geopandas as gpd
import numpy as np
import pandas as pd
from pyproj import CRS, Transformer
import pytz
from scipy.interpolate import interp1d
from shapely import ops
from shapely.geometry import Point


from pyschism import dates


logger = logging.getLogger(__name__)


class SourceSinkDataset:
    def __init__(self, data):
        self._data = data

    def __len__(self):
        return len(self.data)

    def __str__(self):
        return str(self.data)

    @property
    def elements(self):
        if not hasattr(self, "_elements"):
            unique_elements = set()
            for elements in self.data.values():
                for element_id in elements.keys():
                    unique_elements.add(element_id)
            self._elements = list(
                map(str, sorted(list(map(int, unique_elements)))))
        return self._elements

    @property
    def timevector(self):
        if not hasattr(self, '_timevector'):
            self._timevector = list(sorted(self.data.keys()))
        return self._timevector

    @property
    def df(self):
        if not hasattr(self, "_df"):
            data = []
            for time, edata in self.data.items():
                for eid, _edata in edata.items():
                    data.append({"time": time, "element_id": eid, **_edata})
            self._df = pd.DataFrame(data)
        return self._df

    @property
    def data(self):
        return self._data


class Sources(SourceSinkDataset):
    def __init__(self, data):
        for time, edata in data.items():
            assert isinstance(time, datetime)
            for eid, datapoint in edata.items():
                assert (
                    datapoint["flow"] >= 0.0
                ), f"Invalid source point for element_id={eid} during time {str(time)}. "
                f'Sources must be >= 0 but got value of {datapoint["flow"]}.'
                assert isinstance(datapoint["temperature"], float)
                assert isinstance(datapoint["salinity"], float)
        super().__init__(data)


class Sinks(SourceSinkDataset):
    def __init__(self, data):
        for time, edata in data.items():
            assert isinstance(time, datetime)
            for eid, datapoint in edata.items():
                assert (
                    datapoint["flow"] <= 0.0
                ), f"Invalid sink point for element_id={eid} during time {str(time)}. "
                f'Sinks must be <= 0 but got value of {datapoint["flow"]}.'
        super().__init__(data)


class TimeHistoryFile(ABC):
    def __init__(self, source_sink: SourceSinkDataset, start_date, rnday, filename):
        self.dataset = source_sink
        self.filename = filename
        self.start_date = start_date
        self.rnday = rnday

    def __str__(self):
        logger.info(
            f'Generate {self.__class__.__name__.lower()} time history string.')
        start = datetime.now()
        # build ts matrix
        ts_matrix = np.full((len(self.dataset.timevector),
                            len(self.dataset.elements)), np.nan)

        for i, element_id in enumerate(self.dataset.elements):
            ts_matrix[:, i] = self.get_element_timeseries(element_id)

        if np.any(np.any(np.isnan(ts_matrix), axis=0)):
            # handle irregular datasets (untested)
            for column_index in np.where(np.any(np.isnan(ts_matrix), axis=0)):
                finite = np.where(np.isfinite(ts_matrix[:, column_index]))[0]
                ts_matrix[:finite[0], column_index] = 0.
                ts_matrix[finite[-1]:, column_index] = 0.
                fit = interp1d(np.array(self.dataset.timevector)[
                               finite], ts_matrix[finite, column_index])
                non_finite = np.where(np.isfinite(
                    ts_matrix[:, column_index]))[0]
                ts_matrix[non_finite, column_index] = fit(
                    np.array(self.dataset.timevector)[non_finite])

        data = []
        for i, row in enumerate(ts_matrix):
            relative_time = (
                self.dataset.timevector[i] - self.start_date).total_seconds()
            if relative_time < 0:
                continue
            data.append(" ".join([
                f"{relative_time:G}",
                *[f'{x:.4e}' for x in row]
            ]))
        logger.info(
            f'Generate time history string took {datetime.now() - start}.')
        return "\n".join(data)

    def get_element_timeseries(self, element_id):
        values = []
        for time in self.dataset.timevector:
            values.append(self.dataset.data[time].get(
                element_id, {}).get('flow', np.nan))
        return np.array(values)

    def write(self, path: Union[str, os.PathLike], overwrite: bool = False):
        path = pathlib.Path(path)
        if (path / self.filename).exists() and overwrite is not True:
            raise IOError("File exists and overwrite is not True.")
        with open(path / self.filename, "w") as f:
            f.write(str(self))


class Vsource(TimeHistoryFile):
    def __init__(self, sources: Sources, start_date, rnday, filename="vsource.th"):
        super().__init__(sources, start_date, rnday, filename)


class Msource(TimeHistoryFile):
    def __init__(self, sources, start_date, rnday, filename="msource.th"):
        super().__init__(sources, start_date, rnday, filename)

    def __str__(self):
        data = []
        for i, time in enumerate(self.dataset.timevector):
            relative_time = (time - self.start_date).total_seconds()
            if relative_time < 0:
                continue
            line = [f"{relative_time:G}"]
            for element_id in self.dataset.elements:
                temperature = (
                    self.dataset.data[time].get(
                        element_id, {}).get("temperature", -9999.0)
                )
                line.append(f"{temperature: .4e}")
            for element_id in self.dataset.elements:
                salinity = (
                    self.dataset.data[time].get(
                        element_id, {}).get("salinity", -9999.0)
                )
                line.append(f"{salinity: .4e}")
            data.append(" ".join(line))
        return "\n".join(data)

    def get_element_timeseries(self, element_id):
        temp = []
        salt = []
        for time in self.dataset.timevector:
            temp.append(self.dataset.data[time].get(
                element_id, {}).get('temperature', -9999.0))
            salt.append(self.dataset.data[time].get(
                element_id, {}).get('salinity', -9999.0))
        return np.array(temp), np.array(salt)


class Vsink(TimeHistoryFile):
    def __init__(self, sinks: Sinks, start_date, rnday, filename="vsink.th"):
        super().__init__(sinks, start_date, rnday, filename)


class SourceSinkWriter:
    def __init__(self, sources: Sources, sinks: Sinks, filename="source_sink.in"):
        self.sources = sources
        self.sinks = sinks
        self.filename = filename

    def __str__(self):
        logger.info('Generate source_sink string')
        start = datetime.now()
        source_id = self.sources.elements
        data = []
        data.append(f"{len(source_id)}")
        for element_id in source_id:
            data.append(f"{element_id}")
        data.append("")
        sink_id = self.sinks.elements
        data.append(f"{len(sink_id)}")
        for element_id in sink_id:
            data.append(f"{element_id}")
        logger.info(
            f'Generate source_sink string took {datetime.now() - start}')
        return "\n".join(data)

    def write(self, path: Union[str, os.PathLike], overwrite: bool = False):
        path = pathlib.Path(path)
        if (path / self.filename).exists() and overwrite is not True:
            raise IOError("File exists and overwrite is not True.")
        with open(path / self.filename, "w") as f:
            f.write(str(self))


class SourceSink:
    def __add__(self, other):
        source_sink = SourceSink()
        source_sink.sources = Sources(
            {**self.sources.data, **other.sources.data})
        source_sink.sinks = Sinks({**self.sinks.data, **other.sinks.data})
        return source_sink

    def __len__(self):
        return len(self._data)

    def add_data(
        self,
        time: datetime,
        element_id: str,
        flow: float,
        temperature: float = np.nan,
        salinity: float = np.nan,
    ):

        time = dates.localize_datetime(time).astimezone(pytz.utc)
        data_for_element = self._data.get(time, {}).get("element_id", {})

        # TODO: What happens if we have two different flows that both are
        # assigned to the same element? Example: 100 m^3/s @ 1 psu then
        # another flow on the same element of 1 m^3/s @ 100 psu. How do we
        # combine these on a single element? Flow is just simple summation,

        _tmp = data_for_element.get("temperature", np.nan)
        if not np.isnan(_tmp):
            if _tmp != temperature:
                raise NotImplementedError(
                    "Two different values of temperature for same " "time/element."
                )

        _sal = data_for_element.get("salinity", np.nan)
        if not np.isnan(_sal):
            if _sal != salinity:
                raise NotImplementedError(
                    "Two different values of salinity for same time/element."
                )

        self._data.setdefault(time, {}).setdefault(element_id, {}).update(
            {
                "flow": np.nansum([data_for_element.get("flow", np.nan), flow]),
                "temperature": temperature,
                "salinity": salinity,
            }
        )
        if hasattr(self, "_df"):
            del self._df

    def get_element_timeseries(self, element_id):
        data = {}
        element_data = self.df[(self.df["element_id"] == element_id)]
        for row in element_data.sort_values(by=["time"]).itertuples():
            data.setdefault(row.time, {}).update(
                {
                    "flow": row.flow,
                    "temperature": row.temperature,
                    "salinity": row.salinity,
                }
            )
        return data

    def remove_element_timeseries(self, element_id):
        for time in self._data:
            self._data[time].pop(element_id)
        if hasattr(self, "_df"):
            del self._df

    def aggregate_by_radius(self, hgrid, radius):

        logger.info("Begin aggregate_by_radius...")
        start = datetime.now()
        # --- Generate aggregation mapping
        # gather extreme values
        source_max = {element_id: -float("inf")
                      for element_id in self.sources.elements}
        for element_data in self.sources.data.values():
            for element_id, data in element_data.items():
                source_max[element_id] = np.max(
                    [source_max[element_id], data["flow"]])

        sink_max = {element_id: float("inf")
                    for element_id in self.sinks.elements}
        for element_data in self.sinks.data.values():
            for element_id, data in element_data.items():
                sink_max[element_id] = np.min(
                    [sink_max[element_id], data["flow"]])

        aggregate_gdf = []
        for element_id, maxflow in {**source_max, **sink_max}.items():
            element_index = hgrid.elements.get_index_by_id(element_id)
            aggregate_gdf.append(
                {
                    "element_id": element_id,
                    "geometry": hgrid.elements.gdf.iloc[element_index].geometry,
                    "maxflow": maxflow,
                }
            )
        aggregate_gdf = gpd.GeoDataFrame(aggregate_gdf, crs=hgrid.crs).sort_values(
            by="maxflow", key=abs, ascending=False
        )

        aggregation_mapping = {}
        for row in aggregate_gdf.itertuples():
            if row.element_id in aggregation_mapping:
                continue
            aggregation_mapping[row.element_id] = row.element_id
            possible_sources = aggregate_gdf.loc[
                aggregate_gdf.index.difference(
                    np.where(
                        aggregate_gdf["element_id"].isin(
                            list(aggregation_mapping))
                    )
                )
            ]
            circle = get_circle_of_radius(
                row.geometry.centroid.x, row.geometry.centroid.y, radius
            )
            sources_in_circle = possible_sources.loc[possible_sources.within(
                circle)]
            for row_in_circle in sources_in_circle.itertuples():
                aggregation_mapping[row_in_circle.element_id] = row.element_id

        # --- move data from one element to the other
        for current, target in aggregation_mapping.items():
            for time, data in self.get_element_timeseries(current).items():
                self._data[time][target]["flow"] = (
                    self._data[time][current]["flow"] +
                    self._data[time][target]["flow"]
                )

        for current, target in aggregation_mapping.items():
            if current != target:
                self.remove_element_timeseries(current)

        if hasattr(self, "_sources"):
            del self._sources

        if hasattr(self, "_sinks"):
            del self._sinks

        if hasattr(self, "_df"):
            del self._df

        logger.info(f"aggregate_by_radius took {datetime.now()-start}...")

    @staticmethod
    def open(source_sink, vsource, vsink, msource, start_date=None):
        raise NotImplementedError

    def write(
        self,
        path: Union[str, os.PathLike],
        overwrite: bool = False,
        msource: Union[str, bool] = True,
        vsource: Union[str, bool] = True,
        vsink: Union[str, bool] = True,
        source_sink: Union[str, bool] = True,
    ):

        path = pathlib.Path(path)
        path.mkdir(exist_ok=overwrite, parents=True)

        # unpack sources, sinks
        sources, sinks = self.sources, self.sinks

        # write source sink
        if source_sink is True:
            fname = "source_sink.in"
        elif isinstance(source_sink, str):
            fname = source_sink
        if source_sink is not False:
            SourceSinkWriter(sources, sinks, fname).write(path, overwrite)

        if vsource is True:
            fname = "vsource.th"
        elif isinstance(vsource, str):
            fname = vsource
        if vsource is not False:
            Vsource(sources, self.start_date, self.rnday,
                    fname).write(path, overwrite)

        if msource is True:
            fname = "msource.th"
        elif isinstance(msource, str):
            fname = msource
        if msource is not False:
            Msource(sources, self.start_date, self.rnday,
                    fname).write(path, overwrite)

        if vsink is True:
            fname = "vsink.th"
        elif isinstance(vsink, str):
            fname = vsink
        if vsink is not False:
            Vsink(sinks, self.start_date, self.rnday,
                  fname).write(path, overwrite)

    @property
    def sources(self):
        if not hasattr(self, "_sources"):
            sources = {}
            for element_id in list(
                map(str, list(sorted(map(int, self.df.element_id.unique()))))
            ):
                element_data = self.df[(self.df["element_id"] == element_id)]
                flow_data = element_data["flow"]
                if np.all(flow_data > 0.0):
                    # TODO:  Are irregular timeseries allowed?
                    # if not, we need an interpolator here.
                    for row in element_data.sort_values(by=["time"]).itertuples():
                        sources.setdefault(row.time, {})[element_id] = {
                            "flow": row.flow,
                            "temperature": row.temperature,
                            "salinity": row.salinity,
                        }

                # handle elements that are both sources and sinks
                elif not np.all(flow_data < 0) and np.any(flow_data > 0.0):
                    for row in element_data.sort_values(by=["time"]).itertuples():
                        flow = row.flow if row.flow >= 0.0 else 0.0
                        sources.setdefault(row.time, {})[element_id] = {
                            "flow": flow,
                            "temperature": row.temperature,
                            "salinity": row.salinity,
                        }
            self._sources = Sources(sources)
        return self._sources

    @property
    def sinks(self):
        if not hasattr(self, "_sinks"):
            sinks = {}
            for element_id in list(
                map(str, list(sorted(map(int, self.df.element_id.unique()))))
            ):
                element_data = self.df[(self.df["element_id"] == element_id)]
                flow_data = element_data["flow"]
                if np.all(flow_data < 0.0):
                    # TODO:  Are irregular timeseries allowed?
                    # if not, we need an interpolator here.
                    for row in element_data.sort_values(by=["time"]).itertuples():
                        sinks.setdefault(row.time, {})[element_id] = {
                            "flow": row.flow}
                # handle elements that are both sources and sinks
                elif not np.all(flow_data > 0.0) and np.any(flow_data < 0.0):
                    for row in element_data.sort_values(by=["time"]).itertuples():
                        sinks.setdefault(row.time, {})[element_id] = {
                            "flow": row.flow if row.flow <= 0.0 else 0.0
                        }
            self._sinks = Sinks(sinks)
        return self._sinks

    @property
    def start_date(self):
        if not hasattr(self, "_start_date"):
            return self.df.time.min()
        elif self._start_date is None:
            return self.df.time.min()
        return self._start_date

    @start_date.setter
    def start_date(self, start_date):
        self._start_date = start_date
        if start_date is not None:
            self._start_date = dates.localize_datetime(
                start_date).astimezone(pytz.utc)
        return self._start_date

    @property
    def rnday(self):
        if not hasattr(self, "_rnday"):
            return self.df.time.max() - self.df.time.min()
        if self._rnday is None:
            return self.df.time.max() - self.df.time.min()
        return self._rnday

    @rnday.setter
    def rnday(self, rnday):
        self._rnday = rnday
        if rnday is not None:
            self._rnday = (
                rnday if isinstance(
                    rnday, timedelta) else timedelta(days=rnday)
            )
        return self._rnday

    @property
    def data(self):
        return self._data

    @property
    def df(self):
        if not hasattr(self, "_df"):
            _data = []
            for time, element_data in self._data.items():
                for element_id, data in element_data.items():
                    _data.append(
                        {"time": time, "element_id": element_id, **data})
            self._df = pd.DataFrame(_data)
        return self._df


@lru_cache(maxsize=None)
def get_circle_of_radius(lon, lat, radius):
    wgs84 = CRS.from_user_input("+proj=longlat +datum=WGS84 +no_defs")
    aeqd = CRS.from_user_input(
        "+proj=aeqd +R=6371000 +units=m " f"+lat_0={lat} +lon_0={lon}"
    )
    wgs84_to_aeqd = Transformer.from_crs(wgs84, aeqd, always_xy=True).transform
    aeqd_to_wgs84 = Transformer.from_crs(aeqd, wgs84, always_xy=True).transform
    center = Point(float(lon), float(lat))
    point_transformed = ops.transform(wgs84_to_aeqd, center)
    return ops.transform(aeqd_to_wgs84, point_transformed.buffer(radius))



import pathlib

from pyschism.forcing.source_sink.nwm import NationalWaterModel, logger as nwm_logger, NWMElementPairings
from pyschism.forcing.source_sink.base import logger as source_sink_base_logger
from pyschism.mesh import Hgrid


for each file:
    my_idxs.append()

pool.starmap (
    func,
    [(nc, my_idxs[i]) for enumerate(each_file)]
)

def test_nwm():

    hgrid = Hgrid.open('https://raw.githubusercontent.com/geomesh/test-data/main/NWM/hgrid.ll')
    nwm = NationalWaterModel()
    cached_pairings = pathlib.Path('pairings.txt')
    nwm.fetch_data(
        hgrid,
        end_date=5.,
        pairings=None if cached_pairings.is_file() is False else NWMElementPairings.from_file(cached_pairings)
        )
    nwm.pairings.make_plot()
    if cached_pairings.is_file() is False:
        nwm.pairings.save('pairings.txt')


if __name__ == "__main__":
    import logging

    logging.basicConfig(
        format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
        force=True,
    )
    logging.captureWarnings(True)

    log_level = logging.DEBUG

    nwm_logger.setLevel(log_level)
    source_sink_base_logger.setLevel(log_level)

    test_nwm()



from abc import abstractmethod
from enum import Enum

# from typing import Union

from pyschism.forcing.bctides.bctypes import Bctype
from pyschism.forcing import hycom

# from pyschism.forcing.bctides.nudge import TempNudge
# from pyschism.mesh.base import Gr3


class Itetype(Bctype):
    @property
    @abstractmethod
    def itetype(self) -> int:
        pass

    @property
    def forcing_digit(self):
        return self.itetype


class UniformTimeHistoryTemperature(Itetype):

    # def __init__(self, time_history: Dict[str, List[float]], tobc: float = 1.):
    #     self.time_history = time_history
    #     super().__init__(tobc)

    @property
    def itetype(self) -> int:
        return 1


class ConstantTemperature(Itetype):

    def __init__(self, value: float, nudging_factor: float):
        self.value = value
        if not (nudging_factor >= 0) and (nudging_factor <= 1):
            raise ValueError(
                    'Argument `nudging_factor` must be >= 0 and <= 1,'
                    f'but got {nudging_factor}.')
        self.nudging_factor = nudging_factor
        super().__init__()

    @property
    def itetype(self) -> int:
        return 2

    def get_boundary_string(self, *args, **kwargs) -> str:
        boundary_string = [
            f'{self.value:0.6f}',
            f'{self.nudging_factor:0.6f}',
        ]
        return '\n'.join(boundary_string)


class TemperatureInitialConditions(Itetype):

    # def __init__(self, tobc: float = 1., tth):
    #     raise NotImplementedError(f'{self.__class__.__name__}')

    @property
    def itetype(self):
        return 3


class SpatiallyVaryingTimeHistoryTemperature(Itetype):
    class BaroclinicDatabases(Enum):
        #RTOFS = hycom.RTOFS
        GOFS = hycom.GOFS

    def __init__(
        self,
        data_source="gofs",
        nudge: bool = True,
        rlmax=1.5,
        rnu_day=0.25,
    ):
        if isinstance(data_source, str):
            data_source = self.BaroclinicDatabases[data_source.upper()].value()
        if not isinstance(data_source, hycom.Hycom):
            raise TypeError(
                "Argument data_source must be of type str or type "
                f"{type(hycom.Hycom)}, not type {type(data_source)}."
            )
        self.data_source = data_source
        self.data_component = data_source.temperature
        self.nudge = bool(nudge)
        self.rlmax = rlmax
        self.rnu_day = rnu_day

    def get_boundary_string(self, *args, **kwargs):
        return "1."

    # def write(self, *args, **kwargs):
    #     if self.nudge is not None:
    #         self.nudge.write(*args, **kwargs)

    @property
    def itetype(self):
        return 4


class Itetype0(Itetype):
    pass


class Itetype1(UniformTimeHistoryTemperature):
    pass


class Itetype2(ConstantTemperature):
    pass


class Itetype3(TemperatureInitialConditions):
    pass


class Itetype4(SpatiallyVaryingTimeHistoryTemperature):
    pass



from time import time
import os
from datetime import datetime, timedelta
import logging
import pathlib
import f90nml

from pyschism.mesh import Hgrid
from pyschism.forcing.bctides import Bctides, iettype, ifltype, isatype, itetype
from pyschism.forcing.source_sink.nwm import NationalWaterModel, NWMElementPairings
from pyschism.forcing.nws import NWS2, GFS, HRRR

if __name__ == "__main__":

    '''
    Assume files are already located in:
    database='fes2014'
        ~/.local/share/fes2014/eastward_velocity/
        ~/.local/share/fes2014/northward_velocity/  
        ~/.local/share/fes2014/ocean_tide_extrapolated/
    database = 'tpxo'
        ~/.local/share/tpxo/
    '''
    #setup logging
    logging.basicConfig(
        format = "[%(asctime)s] %(name)s %(levelname)s: %(message)s",
        force=True,
    )
    logging.getLogger("pyschism").setLevel(logging.DEBUG)

    start_date = datetime(2017, 12, 1)
    rnday = 396
    end_date = start_date + timedelta(days=61)
    outdir = './'
    hgrid = Hgrid.open("./hgrid.gr3", crs="epsg:4326")

    #elevation
    iet3 = iettype.Iettype3(constituents='major', database='fes2014')
    iet4 = iettype.Iettype4()
    iet5 = iettype.Iettype5(iettype3=iet3, iettype4=iet4)

    #velocity
    ifl1 = ifltype.Ifltype1()
    ifl3 = ifltype.Ifltype3(constituents='major', database='fes2014')
    ifl4 = ifltype.Ifltype4()
    ifl5 = ifltype.Ifltype5(ifltype3=ifl3, ifltype4=ifl4)

    #salinity
    isa4 = isatype.Isatype4()
    isa2 = isatype.Isatype2(5, 1)

    #temperature
    ite4 = itetype.Itetype4()
    ite2 = itetype.Itetype2(0, 1)

    bctides=Bctides(
        hgrid, 
        iettype={'1': iet5, '2': iet5}, 
        ifltype={'1': ifl5, '2': ifl5, '3': ifl1}, 
        isatype={'1': isa4, '2':isa4, '3': isa2}, 
        itetype={'1': ite4, '2':ite4, '3': ite2}
    )

    bctides.write(
        outdir, 
        start_date=start_date, 
        end_date=end_date, 
        bctides=True, 
        overwrite=True
    )



import signal
import os
import errno
from functools import wraps
from datetime import datetime, timedelta
from enum import Enum
import pathlib
import tempfile
from typing import Union
import logging

import pytz
from matplotlib.transforms import Bbox
from netCDF4 import Dataset
import numpy as np

from pyschism.enums import GFSProduct
from pyschism.forcing.nws.nws2.sflux import (
    SfluxDataset,
    AirComponent,
    PrcComponent,
    RadComponent,
)
from pyschism.dates import nearest_zulu, localize_datetime, nearest_cycle

BASE_URL = "https://nomads.ncep.noaa.gov/dods"
logger = logging.getLogger(__name__)


class BaseURL(Enum):
    GFS_0P25 = f"{BASE_URL}/gfs_0p25"
    GFS_0P25_1HR = f"{BASE_URL}/gfs_0p25_1hr"
    GFS_0P50 = f"{BASE_URL}/gfs_0p50"
    GFS_1P00 = f"{BASE_URL}/gfs_1p00"

    @classmethod
    def _missing_(self, name):
        raise ValueError(f"{name} is not a valid GFS product.")


class TimeoutError(Exception):
    pass


def timeout(seconds=30, error_message=os.strerror(errno.ETIME)):
    def decorator(func):
        def _handle_timeout(signum, frame):
            raise TimeoutError(error_message)

        def wrapper(*args, **kwargs):
            signal.signal(signal.SIGALRM, _handle_timeout)
            signal.alarm(seconds)
            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result

        return wraps(func)(wrapper)

    return decorator


class GFSInventory:
    def __init__(self, product="gfs_0p25_1hr", start_date=None, rnday=4, bbox=None):
        self.product = (
            GFSProduct(product.upper())
            if not isinstance(product, GFSProduct)
            else product
        )
        self.start_date = (
            nearest_cycle()
            if start_date is None
            else localize_datetime(start_date).astimezone(pytz.utc)
        )
        self.rnday = rnday if isinstance(rnday, timedelta) else timedelta(days=rnday)
        # if self.start_date != nearest_cycle(self.start_date):
        #     raise NotImplementedError(
        #         "Argument start_date is does not align with any GFS cycle " "times."
        #     )
        self.check_dates(self.start_date, self.rnday)
        end_date = (self.start_date + self.rnday).astimezone(pytz.utc)
        end_date.replace(tzinfo=None)
        self._files = {
            _: None
            for _ in np.arange(
                self.start_date,
                end_date + self.output_interval,
                self.output_interval,
            ).astype(datetime)
        }
        
        nearest_end_date = nearest_cycle(datetime.utcnow(), period=6)
        nearest_end_date.replace(tzinfo=None)
        for dt in self.nearest_zulus:
            if None not in list(self._files.values()):
                break
            base_url = (
                BASE_URL
                + f"/{self.product.value}"
                + f'/gfs{nearest_zulu(dt).strftime("%Y%m%d")}'
            )
            for cycle in reversed(range(0, 24, 6)):
                if np.datetime64(dt + timedelta(hours=cycle)) > np.datetime64(
                    nearest_end_date
                ):
                    continue
                test_url = f"{base_url}/" + f"{self.product.name.lower()}_{cycle:02d}z"
                nc = self.fetch_nc_by_url(test_url)
                if nc is None:
                    continue
                file_dates = self.get_nc_datevector(nc)
                for _datetime in reversed(list(self._files.keys())):
                    if _datetime in file_dates:
                        if self._files[_datetime] is None:
                            self._files[_datetime] = nc
                if not any(nc is None for nc in self._files.values()):
                    break

        missing_records = [dt for dt, nc in self._files.items() if nc is None]
        if len(missing_records) > 0:
            raise ValueError(f"No GFS data for dates: {missing_records}.")

        self._bbox = self._modified_bbox(bbox)

    def check_dates(self, start_date, rnday):

        end_date = (start_date + rnday).astimezone(pytz.utc)
        end_date.replace(tzinfo=None)
        nc = None
        i = 0
        while nc is None:

            dt = nearest_cycle(
                nearest_cycle(datetime.utcnow()) - (i * self.output_interval)
            )
            cycle = (6 * np.floor(dt.hour / 6)) % 24
            base_url = (
                BASE_URL + f"/{self.product.value}" + f'/gfs{dt.strftime("%Y%m%d")}'
            )
            test_url = f"{base_url}/" + f"{self.product.name.lower()}_{int(cycle):02d}z"
            nc = self.fetch_nc_by_url(test_url)
            i += 1
        datevec = self.get_nc_datevector(nc)
        # print(np.datetime64(np.max(datevec)))
        # print(np.datetime64(end_date - self.output_interval))
        if np.datetime64(np.max(datevec)) < np.datetime64(
            end_date - 2 * self.output_interval
        ):
            raise IOError(
                f"Requested end date at {end_date - self.output_interval} is larger than the max allowed GFS end_date "
                f" {np.max(datevec) - self.output_interval}"
            )

    @staticmethod
    def fetch_nc_by_url(url):
        try:
            logger.info(f"Checking url: {url}")

            @timeout()
            def get_netcdf_timeout():
                return Dataset(url)

            nc = get_netcdf_timeout()
            logger.info("Success!")
            return nc
        except OSError as e:
            if e.errno == -70:
                print()
                logger.info("URL not yet in server.")
                return None
            elif e.errno == -73:
                nc = False

                @timeout()
                def retry():
                    try:
                        return Dataset(url)
                    except Exception:
                        return False

                while not isinstance(nc, Dataset):
                    logger.info("retrying...")
                    nc = retry()

            elif e.errno == -72:
                return None

            else:
                raise e

    def put_sflux_field(self, gfs_varname: str, dst: Dataset, sflux_varname: str):

        lon_idxs, lat_idxs = self._modified_bbox_indexes(self._bbox)
        for i, (dt, nc) in enumerate(self._files.items()):
            time_index = self.get_nc_time_index(nc, dt)
            if time_index == 0:
                time_index += 1
            logger.info(
                f"Putting GFS field {gfs_varname} for time {dt} as "
                f"{sflux_varname} from file "
                f'{nc.filepath().replace(f"{BASE_URL}/", "")}[{time_index}].'
            )
            # TODO: exponential time backoff retry
            def put_nc_field():
                try:
                    dst[sflux_varname][i, :, :] = nc.variables[gfs_varname][
                        time_index, lat_idxs, lon_idxs
                    ]
                    dst.sync()
                    if np.any(dst[sflux_varname][i, :, :].mask) is True:
                        raise Exception(
                            f"Junk value in nc_field {gfs_varname}, index {time_index}"
                        )
                    return True
                except RuntimeError:
                    return False

            success = False
            while success is False:
                success = put_nc_field()

    def get_nc_time_index(self, nc, dt):
        return np.where(np.in1d(self.get_nc_datevector(nc), [dt]))[0][0]

    def get_nc_datevector(self, nc):
        try:
            base_date = localize_datetime(
                datetime.strptime(nc["time"].minimum.split("z")[-1], "%d%b%Y")
            ) + timedelta(hours=float(nc["time"].minimum.split("z")[0]))
            return np.arange(
                base_date,
                base_date + timedelta(len(nc["time"][:])),
                self.output_interval,
            ).astype(datetime)
        except RuntimeError:
            return self.get_nc_datevector(nc)

    def get_sflux_timevector(self):
        timevec = list(self._files.keys())
        _nearest_zulu = nearest_zulu(np.min(timevec))
        return [
            (localize_datetime(x) - _nearest_zulu) / timedelta(days=1) for x in timevec
        ]

    def xy_grid(self):
        lon_idxs, lat_idxs = self._modified_bbox_indexes(self._bbox)
        lon = []
        for x in self.lon[lon_idxs]:
            if x > 180:
                lon.append(x - 360)
            else:
                lon.append(x)
        return np.meshgrid(np.array(lon), self.lat[lat_idxs])

    @property
    def nearest_zulu(self):
        if not hasattr(self, "_nearest_zulu"):
            self._nearest_zulu = nearest_zulu()
        return self._nearest_zulu

    @property
    def nearest_zulus(self):
        return np.arange(
            self.nearest_zulu,
            self.nearest_zulu - timedelta(days=10),
            -timedelta(days=1),
        ).astype(datetime)

    @property
    def output_interval(self):
        if self.product == GFSProduct.GFS_0P25_1HR:
            return timedelta(hours=1)
        return timedelta(hours=6)

    @property
    def lon(self):
        if not hasattr(self, "_lon"):
            nc = self._files[list(self._files.keys())[0]]
            self._lon = nc.variables["lon"][:]
            if not hasattr(self, "_lat"):
                self._lat = nc.variables["lat"][:]
        return self._lon

    @property
    def lat(self):
        if not hasattr(self, "_lat"):
            nc = self._files[list(self._files.keys())[0]]
            self._lat = nc.variables["lat"][:]
            if not hasattr(self, "_lon"):
                self._lon = nc.variables["lon"][:]
        return self._lat

    def _modified_bbox(self, bbox=None):
        if bbox is None:
            return Bbox.from_extents(0, -90, 360, 90)
        else:
            xmin = bbox.xmin + 360 if bbox.xmin < 0 else bbox.xmin
            xmax = bbox.xmax + 360 if bbox.xmax < 0 else bbox.xmax
            return Bbox.from_extents(xmin, bbox.ymin, xmax, bbox.ymax)

    def _modified_bbox_indexes(self, bbox):
        lat_idxs = np.where((self.lat >= bbox.ymin-1.0) & (self.lat <= bbox.ymax+1.0))[0]
        lon_idxs = np.where((self.lon >= bbox.xmin-1.0) & (self.lon <= bbox.xmax+1.0))[0]
        return lon_idxs, lat_idxs


class GlobalForecastSystem(SfluxDataset):
    def __init__(
        self,
        product: Union[str, GFSProduct] = GFSProduct.GFS_0P25_1HR,
    ):
        self.prmsl_name = "prmslmsl"
        self.spfh_name = "spfh2m"
        self.stmp_name = "tmpsfc"
        self.uwind_name = "ugrd10m"
        self.vwind_name = "vgrd10m"
        self.prate_name = "pratesfc"
        self.dlwrf_name = "dlwrfsfc"
        self.dswrf_name = "dswrfsfc"
        self.product = (
            GFSProduct(product) if not isinstance(product, GFSProduct) else product
        )
        self.air = None
        self.prc = None
        self.rad = None

    def write(
        self,
        outdir: Union[str, os.PathLike],
        level,
        start_date: datetime = None,
        rnday: Union[float, timedelta] = 4,
        air: bool = True,
        prc: bool = True,
        rad: bool = True,
        bbox: Bbox = None,
        overwrite: bool = False,
    ):
        """Fetches GFS data from NOMADS server."""
        logger.info("Fetching GFS data.")
        self.start_date = (
            nearest_cycle()
            if start_date is None
            else localize_datetime(start_date).astimezone(pytz.utc)
        )
        self.rnday = rnday if isinstance(rnday, timedelta) else timedelta(days=rnday)

        inventory = GFSInventory(
            self.product, self.start_date, self.rnday + self.output_interval, bbox
        )

        nx_grid, ny_grid = inventory.xy_grid()
        if air is True:
            with Dataset(
                self.tmpdir / f"air_{inventory.product.value}_"
                f"{str(self.start_date)}.nc",
                "w",
                format="NETCDF3_CLASSIC",
            ) as dst:

                # global attributes
                dst.setncatts({"Conventions": "CF-1.0"})
                # dimensions
                dst.createDimension("nx_grid", nx_grid.shape[1])
                dst.createDimension("ny_grid", ny_grid.shape[0])
                dst.createDimension("time", None)
                # variables
                # lon
                dst.createVariable("lon", "f4", ("ny_grid", "nx_grid"))
                dst["lon"].long_name = "Longitude"
                dst["lon"].standard_name = "longitude"
                dst["lon"].units = "degrees_east"
                dst["lon"][:] = nx_grid
                # lat
                dst.createVariable("lat", "f4", ("ny_grid", "nx_grid"))
                dst["lat"].long_name = "Latitude"
                dst["lat"].standard_name = "latitude"
                dst["lat"].units = "degrees_north"
                dst["lat"][:] = ny_grid
                # time
                dst.createVariable("time", "f4", ("time",))
                dst["time"].long_name = "Time"
                dst["time"].standard_name = "time"
                date = nearest_zulu(self.start_date)
                dst["time"].units = (
                    f"days since {date.year}-{date.month}"
                    f"-{date.day} 00:00"
                    f"{date.tzinfo}"
                )
                dst["time"].base_date = (date.year, date.month, date.day, 0)
                dst["time"][:] = inventory.get_sflux_timevector()

                for var in AirComponent.var_types:
                    dst.createVariable(var, "f4", ("time", "ny_grid", "nx_grid"))
                    logger.info(f"Put field {var}")
                    inventory.put_sflux_field(getattr(self, f"{var}_name"), dst, var)

                # prmsl
                dst["prmsl"].long_name = "Pressure reduced to MSL"
                dst["prmsl"].standard_name = "air_pressure_at_sea_level"
                dst["prmsl"].units = "Pa"

                # spfh
                dst["spfh"].long_name = "Surface Specific Humidity " "(2m AGL)"
                dst["spfh"].standard_name = "specific_humidity"
                dst["spfh"].units = "1"

                # stmp
                dst["stmp"].long_name = "Surface Air Temperature (2m AGL)"
                dst["stmp"].standard_name = "air_temperature"
                dst["stmp"].units = "K"

                # uwind
                dst["uwind"].long_name = "Surface Eastward Air Velocity " "(10m AGL)"
                dst["uwind"].standard_name = "eastward_wind"
                dst["uwind"].units = "m/s"

                # vwind
                dst["vwind"].long_name = "Surface Northward Air Velocity " "(10m AGL)"
                dst["vwind"].standard_name = "northward_wind"
                dst["vwind"].units = "m/s"

        if prc is True:
            with Dataset(
                self.tmpdir / f"prc_{inventory.product.value}_"
                f"{str(self.start_date)}.nc",
                "w",
                format="NETCDF3_CLASSIC",
            ) as dst:

                # global attributes
                dst.setncatts({"Conventions": "CF-1.0"})
                # dimensions
                dst.createDimension("nx_grid", nx_grid.shape[1])
                dst.createDimension("ny_grid", ny_grid.shape[0])
                dst.createDimension("time", None)
                # lon
                dst.createVariable("lon", "f4", ("ny_grid", "nx_grid"))
                dst["lon"].long_name = "Longitude"
                dst["lon"].standard_name = "longitude"
                dst["lon"].units = "degrees_east"
                dst["lon"][:] = nx_grid
                # lat
                dst.createVariable("lat", "f4", ("ny_grid", "nx_grid"))
                dst["lat"].long_name = "Latitude"
                dst["lat"].standard_name = "latitude"
                dst["lat"].units = "degrees_north"
                dst["lat"][:] = ny_grid
                # time
                dst.createVariable("time", "f4", ("time",))
                dst["time"].long_name = "Time"
                dst["time"].standard_name = "time"
                date = nearest_zulu(self.start_date)
                dst["time"].units = (
                    f"days since {date.year}-{date.month}"
                    f"-{date.day} 00:00"
                    f"{date.tzinfo}"
                )
                dst["time"].base_date = (date.year, date.month, date.day, 0)
                dst["time"][:] = inventory.get_sflux_timevector()

                for var in PrcComponent.var_types:
                    dst.createVariable(var, float, ("time", "ny_grid", "nx_grid"))
                    logger.info(f"Put field {var}")
                    inventory.put_sflux_field(getattr(self, f"{var}_name"), dst, var)
                # prate
                dst["prate"].long_name = "Surface Precipitation Rate"
                dst["prate"].standard_name = "air_pressure_at_sea_level"
                dst["prate"].units = "kg/m^2/s"

        if rad is True:
            with Dataset(
                self.tmpdir / f"rad_{inventory.product.value}_"
                f"{str(self.start_date)}.nc",
                "w",
                format="NETCDF3_CLASSIC",
            ) as dst:
                # global attributes
                dst.setncatts({"Conventions": "CF-1.0"})
                # dimensions
                dst.createDimension("nx_grid", nx_grid.shape[1])
                dst.createDimension("ny_grid", ny_grid.shape[0])
                dst.createDimension("time", None)
                # lon
                dst.createVariable("lon", "f4", ("ny_grid", "nx_grid"))
                dst["lon"].long_name = "Longitude"
                dst["lon"].standard_name = "longitude"
                dst["lon"].units = "degrees_east"
                dst["lon"][:] = nx_grid
                # lat
                dst.createVariable("lat", "f4", ("ny_grid", "nx_grid"))
                dst["lat"].long_name = "Latitude"
                dst["lat"].standard_name = "latitude"
                dst["lat"].units = "degrees_north"
                dst["lat"][:] = ny_grid
                # time
                dst.createVariable("time", "f4", ("time",))
                dst["time"].long_name = "Time"
                dst["time"].standard_name = "time"
                date = nearest_zulu(self.start_date)
                dst["time"].units = (
                    f"days since {date.year}-{date.month}"
                    f"-{date.day} 00:00"
                    f"{date.tzinfo}"
                )
                dst["time"].base_date = (date.year, date.month, date.day, 0)
                dst["time"][:] = inventory.get_sflux_timevector()

                for var in RadComponent.var_types:
                    dst.createVariable(var, float, ("time", "ny_grid", "nx_grid"))
                    logger.info(f"Put field {var}")
                    inventory.put_sflux_field(getattr(self, f"{var}_name"), dst, var)

                # dlwrf
                dst["dlwrf"].long_name = "Downward Long Wave Radiation " "Flux"
                dst["dlwrf"].standard_name = (
                    "surface_downwelling_" "longwave_flux_in_air"
                )
                dst["dlwrf"].units = "W/m^2"

                # dswrf
                dst["dswrf"].long_name = "Downward Short Wave Radiation " "Flux"
                dst["dswrf"].standard_name = (
                    "surface_downwelling_" "shortwave_flux_in_air"
                )
                dst["dswrf"].units = "W/m^2"

        self.resource = list(self.tmpdir.glob("*.nc"))
        if air is True:
            self.air = AirComponent(self.fields)

        if prc is True:
            self.prc = PrcComponent(self.fields)

        if rad is True:
            self.rad = RadComponent(self.fields)

        super().write(
            outdir,
            level,
            overwrite=overwrite,
            start_date=start_date,
            rnday=rnday,
            air=air,
            rad=rad,
            prc=prc,
        )
        del self._tmpdir

    @property
    def tmpdir(self):
        if not hasattr(self, "_tmpdir"):
            self._tmpdir = tempfile.TemporaryDirectory()
        return pathlib.Path(self._tmpdir.name)

    @property
    def output_interval(self):
        if self.product == GFSProduct.GFS_0P25_1HR:
            return timedelta(hours=1)
        return timedelta(hours=6)



from abc import ABC, abstractmethod
from datetime import datetime, timedelta
import logging
import os
import pathlib
from typing import Union
import warnings

import cf
import cftime
from netCDF4 import Dataset
import numpy as np
import pytz

from pyschism.dates import localize_datetime


_logger = logging.getLogger(__name__)


class FieldsDescriptor:

    def __set__(self, obj, fields: cf.FieldList):
        if len(fields) == 0:
            warnings.warn(f"No data with name {obj.name} exists on the input "
                          'dataset.')
        ordered_indexes = np.argsort(
            [field.construct('time').reference_datetime for field
             in fields])
        fields = cf.FieldList([fields[idx] for idx in ordered_indexes])
        obj.__dict__[obj.type] = fields

    def __get__(self, obj, val):
        return obj.__dict__[obj.type]


class Nxgrids:

    def __set__(self, obj, fields: cf.FieldList):
        if len(fields) == 0:
            raise ValueError("No data with name 'lon' exists on the input "
                             'dataset.')
        ordered_indexes = np.argsort(
            [field.construct('time').reference_datetime for field
             in fields])
        fields = cf.FieldList([fields[idx] for idx in ordered_indexes])
        obj.__dict__['nx_grids'] = fields

    def __get__(self, obj, val):
        return obj.__dict__['nx_grids']


class Nygrids:

    def __set__(self, obj, fields: cf.FieldList):
        if len(fields) == 0:
            raise ValueError("No data with name 'lat' exists on the input "
                             'dataset.')
        ordered_indexes = np.argsort(
            [field.construct('time').reference_datetime for field
             in fields])
        fields = cf.FieldList([fields[idx] for idx in ordered_indexes])
        obj.__dict__['ny_grids'] = fields

    def __get__(self, obj, val):
        return obj.__dict__['ny_grids']


class DatetimeArrays:

    def __get__(self, obj, val):
        for field in obj.fields:
            dt_array = []
            for dt in field.construct('time').datetime_array:
                if not isinstance(dt, cftime.DatetimeGregorian):
                    continue
                dt_array.append(pytz.timezone('UTC').localize(
                    datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute)))
            yield dt_array


class DatetimeArray:

    def __get__(self, obj, val):
        dt_array = set()
        for dt_array_x in obj.datetime_arrays:
            for dt in dt_array_x:
                dt_array.add(dt)
        return [dt for dt in sorted(dt_array)]


class ReferenceDatetimes:

    def __get__(self, obj, val):
        for field in obj.fields:
            dt = field.construct('time').reference_datetime
            yield pytz.timezone('UTC').localize(
                datetime(dt.year, dt.month, dt.day, dt.hour, dt.minute)
                )


class Variable:

    fields = FieldsDescriptor()
    reference_datetimes = ReferenceDatetimes()
    datetime_arrays = DatetimeArrays()
    datetime_array = DatetimeArray()

    def __init__(self, fields, var_type, name, long_name,
                 standard_name, units):
        self.type = var_type
        self.name = name
        self.nx_grids = fields.select_by_ncvar('lon')
        self.ny_grids = fields.select_by_ncvar('lat')
        self.fields = fields.select_by_ncvar(self.name)
        self.long_name = long_name
        self.standard_name = standard_name
        self.units = units

    def get_fields(self, start_date: datetime = None,
                   rnday: Union[float, int, timedelta] = None,
                   ) -> cf.FieldList:
        if start_date is None:
            start_date = self.datetime_array[0]

        if start_date < np.min(self.datetime_array) or \
                start_date > np.max(self.datetime_array):
            raise ValueError(f'Requested start date {start_date} is out of '
                             f'range with {np.min(self.datetime_array)} and '
                             f'{np.max(self.datetime_array)}')
        if rnday is None:
            rnday = np.max(self.datetime_array[:-2]) - start_date

        elif isinstance(rnday, (int, float)):
            rnday = timedelta(days=rnday)

        end_date = start_date + rnday  # type: ignore[operator]

        #if end_date <= np.min(self.datetime_array) or \
        #        end_date >= np.max(self.datetime_array):
        #    rnday = end_date - start_date
        #    raise ValueError(f'Requested rnday {rnday} has an end date of '
        #                     f'{end_date} which is out of range with '
        #                     f'start_date={np.min(self.datetime_array)} and '
        #                     f'end_date={np.max(self.datetime_array)}')
        fields = []
        for i, datetime_array in reversed(
                list(enumerate(self.datetime_arrays))):
            field_start_date = np.min(datetime_array)
            if field_start_date >= start_date:
                fields.append(self.fields[i])
            else:
                fields.append(self.fields[i])
                break
        return cf.FieldList(fields)

    def animation(
            self,
            save=False,
            fps=3,
            start_frame=0,
            end_frame=-1,
            figsize=None,
            wireframe=False,
            cmap='jet',
            levels=256,
            show=False,
            xmin=None,
            xmax=None,
            ymin=None,
            ymax=None,
            vmin=None,
            vmax=None,

    ):
        from matplotlib.animation import FuncAnimation
        import matplotlib.pyplot as plt
        # self.current_step = start_frame
        fig = plt.figure(figsize)
        ax = fig.add_subplot(111)

        plt.tight_layout(pad=2)

        # triangulation = self.hgrid.triangulation
        # xmin = np.min(self.hgrid.x) if xmin is None else xmin
        # xmax = np.max(self.hgrid.x) if xmax is None else xmax
        # ymin = np.min(self.hgrid.y) if ymin is None else ymin
        # ymax = np.max(self.hgrid.y) if ymax is None else ymax
        # vmin = np.min(self.values) if vmin is None else vmin
        # vmax = np.max(self.values) if vmax is None else vmax

        # triangulation.set_mask(self.hgrid.elements.get_triangulation_mask(self.parent.wetdry_elem.values))
        
        arrays = self.get_fields()[0].array
        def animate(index):
            # self.current_step = index
            _ax = fig.get_axes()
            ax.clear()
            if len(_ax) > 1:
                cax = _ax[1]
                cax.cla()
            else:
                cax = None

            # triangulation.set_mask(self.hgrid.elements.get_triangulation_mask(self.parent.wetdry_elem.values))

            # if wireframe:
            #     ax.triplot(triangulation, color='k', linewidth=0.7)

            # ax.tricontourf(
            #     triangulation,
            #     self.values,
            #     cmap=cmap,
            #     levels=levels,
            #     vmin=vmin,
            #     vmax=vmax
            #     )
            # print(fields[index].array)
            #print(self.datetime_array[index], arrays[index])
            ax.contourf(arrays[index])

            # ax.set_ylim(ymin, ymax, auto=True)
            # ax.set_xlim(xmin, xmax, auto=True)

            ax.set_xlabel('Longitude (°E)')
            ax.set_ylabel('Latitude (°N)')

            ax.set_title(self.datetime_array[index].strftime('%b %d, %Y %H:%M'))
            # m = plt.cm.ScalarMappable(cmap=cmap)
            # m.set_array(self.values[index])
            # m.set_clim(vmin, vmax)
            # cbar = fig.colorbar(
            #     m, cax=cax, format='%.1f',
            #     boundaries=np.linspace(vmin, vmax, levels)
            # )

            # cbar = fig.colorbar(_ax)
            # cbar.ax.set_ylabel(f'{variable} [{unit}]', rotation=90)

        end_frame = end_frame % len(arrays) if end_frame < 0 else end_frame
        start_frame = start_frame % len(arrays) if start_frame < 0 else start_frame
        frames = range(start_frame, end_frame)
        # print(fields.array[start_frame])
        ax.contourf(arrays[start_frame])
        anim = FuncAnimation(
            fig,
            animate,
            frames,
            blit=False
            )

        if save:
            anim.save(
                pathlib.Path(save),
                writer='imagemagick',
                fps=fps
            )

        if show:
            plt.show()

        return anim


class BaseComponent(ABC):

    name: str = ''

    def write(self, outdir: Union[str, os.PathLike], level: int,
              overwrite: bool = False, start_date: datetime = None,
              rnday: Union[float, int, timedelta] = None):

        assert level in [1, 2]
        outdir = pathlib.Path(outdir)

        if start_date is None:
            for vartype in self.var_types:
                variable = getattr(self, vartype)
                if start_date is None:
                    start_date = np.min(variable.datetime_array)

        if start_date is not None:
            # naive condition
            if start_date.tzinfo is None \
                    or start_date.tzinfo.utcoffset(start_date) is None:
                start_date = pytz.timezone('UTC').localize(start_date)
            timezone = start_date.tzinfo

        stacks = []
        for i, field in enumerate(
            getattr(
                self,
                self.var_types[0]
                ).get_fields(start_date, rnday)
                ):
            stacks.append(f"sflux_{self.name}_{level}.{i+1:04d}.nc")
        for i, filename in enumerate(stacks):
            with Dataset(outdir / filename, 'w',
                         format='NETCDF3_CLASSIC') as dst:
                dst.setncatts({"Conventions": "CF-1.0"})
                # dimensions
                variable = getattr(self, self.var_types[0])
                dst.createDimension('nx_grid',
                                    variable.nx_grids[0].shape[1])
                dst.createDimension('ny_grid',
                                    variable.ny_grids[0].shape[0])
                dst.createDimension('time', None)
                # variables
                # lon
                dst.createVariable('lon', 'f4', ('ny_grid', 'nx_grid'))
                dst['lon'].long_name = "Longitude"
                dst['lon'].standard_name = "longitude"
                dst['lon'].units = "degrees_east"
                dst['lon'][:] = variable.nx_grids[0]
                # lat
                dst.createVariable('lat', 'f4', ('ny_grid', 'nx_grid'))
                dst['lat'].long_name = "Latitude"
                dst['lat'].standard_name = "latitude"
                dst['lat'].units = "degrees_north"
                dst['lat'][:] = variable.ny_grids[0]
                nc_start_date = list(variable.reference_datetimes)[0]
                nc_start_date = nc_start_date.astimezone(timezone)
                dst.createVariable('time', 'f4', ('time',))
                dst['time'].long_name = 'Time'
                dst['time'].standard_name = 'time'
                dst['time'].units = f'days since {nc_start_date.year}-' \
                                    f'{nc_start_date.month}-'\
                                    f'{nc_start_date.day} '\
                                    '00:00:00+' \
                                    f'{nc_start_date.tzinfo}'
                dst['time'].base_date = (
                    nc_start_date.year,
                    nc_start_date.month,
                    nc_start_date.day,
                    0)
                dst['time'][:] = [
                    (localize_datetime(x) - nc_start_date) / timedelta(days=1)
                    for x in variable.datetime_array]
                for vartype in self.var_types:
                    variable = getattr(self, vartype)
                    dst.createVariable(
                        variable.name, 'f4', ('time', 'ny_grid', 'nx_grid'))
                    for field in variable.get_fields(start_date, rnday):
                        dst[variable.name][:] = field
                        setattr(
                            dst[variable.name], "long_name",
                            variable.long_name)
                        setattr(
                            dst[variable.name], "standard_name",
                            variable.standard_name)
                        setattr(
                            dst[variable.name], "units",
                            variable.units)

    @property
    @abstractmethod
    def var_types(self):
        """
        air: ['prmsl', 'spfh', 'stmp', 'uwind', 'vwind']
        prc: ['prate']
        rad: ['dlwrf', 'dswrf']
        """

    @property
    def timevector(self):
        return getattr(self, self.var_types[0]).datetime_array


class AirComponent(BaseComponent):

    name = 'air'
    var_types = ['prmsl', 'spfh', 'stmp', 'uwind', 'vwind']

    def __init__(self, fields: cf.FieldList, prmsl_name='prmsl',
                 spfh_name='spfh', stmp_name='stmp', uwind_name='uwind',
                 vwind_name='vwind'):
        self.prmsl = Variable(fields, 'prmsl', prmsl_name,
                              "Pressure reduced to MSL",
                              "air_pressure_at_sea_level", "Pa")
        self.spfh = Variable(fields, 'spfh', spfh_name,
                             "Surface Specific Humidity (2m AGL)",
                             "specific_humidity", '1')
        self.stmp = Variable(fields, 'stmp', stmp_name,
                             "Surface Air Temperature (2m AGL)",
                             "air_temperature", "K")
        self.uwind = Variable(fields, 'uwind', uwind_name,
                              "Surface Eastward Air Velocity (10m AGL)",
                              "eastward_wind", "m/s")
        self.vwind = Variable(fields, 'vwind', vwind_name,
                              "Surface Northward Air Velocity (10m AGL)",
                              "northward_wind", "m/s")


class PrcComponent(BaseComponent):

    name = 'prc'
    var_types = ['prate']

    def __init__(self, fields: cf.FieldList, prate_name='prate'):
        self.prate = Variable(fields, 'prate', prate_name,
                              "Surface Precipitation Rate",
                              "air_pressure_at_sea_level", "kg/m^2/s")


class RadComponent(BaseComponent):

    name = 'rad'
    var_types = ['dlwrf', 'dswrf']

    def __init__(self, fields: cf.FieldList, dlwrf_name='dlwrf',
                 dswrf_name='dswrf'):
        self.dlwrf = Variable(fields, 'dlwrf', dlwrf_name,
                              "Downward Long Wave Radiation Flux",
                              "surface_downwelling_longwave_flux_in_air",
                              "W/m^2")
        self.dswrf = Variable(fields, 'dswrf', dswrf_name,
                              "Downward Short Wave Radiation Flux",
                              "surface_downwelling_shortwave_flux_in_air",
                              "W/m^2")


class SfluxDataset:

    def __init__(self, resource: Union[str, os.PathLike], prmsl_name='prmsl',
                 spfh_name='spfh', stmp_name='stmp', uwind_name='uwind',
                 vwind_name='vwind', prate_name='prate', dlwrf_name='dlwrf',
                 dswrf_name='dswrf'):
        self.resource = resource
        self.air = AirComponent(self.fields, prmsl_name=prmsl_name,
                                spfh_name=spfh_name, stmp_name=stmp_name,
                                uwind_name=uwind_name, vwind_name=vwind_name)
        self.prc = PrcComponent(self.fields, prate_name=prate_name)
        self.rad = RadComponent(self.fields, dlwrf_name=dlwrf_name,
                                dswrf_name=dswrf_name)

    def write(
            self,
            outdir: Union[str, os.PathLike],
            level: int,
            overwrite: bool = False,
            start_date: datetime = None,
            rnday: Union[float, int, timedelta] = None,
            air=True,
            rad=True,
            prc=True,
            bbox=None,
    ):

        outdir = pathlib.Path(outdir)
        if outdir.name != 'sflux':
            outdir /= 'sflux'
        outdir.mkdir(exist_ok=True)

        if hasattr(self, 'air'):
            if self.air is not None:
                if air is True:
                    self.air.write(outdir, level, overwrite, start_date, rnday)

        if hasattr(self, 'prc'):
            if self.prc is not None:
                if rad is True:
                    self.prc.write(outdir, level, overwrite, start_date, rnday)

        if hasattr(self, 'rad'):
            if self.rad is not None:
                if rad is True:
                    self.rad.write(outdir, level, overwrite, start_date, rnday)

    @property
    def timevector(self):
        for attr in ['air', 'prc', 'rad']:
            if hasattr(self, attr):
                return getattr(self, attr).timevector

    @property
    def resource(self):
        return self._resource

    @resource.setter
    def resource(self, resource):
        self._resource = resource
        if hasattr(self, '_fields'):
            del self._fields

    @property
    def fields(self):
        if not hasattr(self, '_fields'):
            fields = cf.read(self.resource, ignore_read_error=True)
            # check lon
            try:
                lon = fields.select_by_ncvar('lon')[0]
            except IndexError:
                raise ValueError(f"Resource {self.resource} does not contain a "
                                 "'lon' variable.")
            if len(lon.get_data_axes()) != 2:
                raise ValueError("'lon' variable must be a 2-dimensional "
                                 "array")
            fnames = lon.get_filenames()
            lons = fields.select_by_ncvar('lon')
            _logger.info(f'fields.select_by_var() returned {lons}')
            for i in range(len(lons) - 1):
                if not (lon.array == lons[i+1].array).all():
                    raise ValueError(
                        "Invalid sflux dataset. Found two different 'lon' "
                        f"fields on files {fnames} and "
                        f'{lons[i+1].get_filenames()}')
            # check lat
            try:
                lat = fields.select_by_ncvar('lat')[0]
            except IndexError:
                raise ValueError(f"Resource {self.resource} does not contain a "
                                 "'lat' variable.")
            if len(lat.get_data_axes()) != 2:
                raise ValueError("'lat' variable must be a 2-dimensional "
                                 "array")
            fnames = lat.get_filenames()
            lats = fields.select_by_ncvar('lat')
            for i in range(len(lats) - 1):
                if not (lat.array == lats[i+1].array).all():
                    raise ValueError(
                        "Invalid sflux dataset. Found two different 'lat' "
                        f"fields on files {fnames} and "
                        f'{lats[i+1].get_filenames()}')
            self._fields = fields
        return self._fields



from pyschism.forcing.source_sink.base import (
    # Hydrology,
    # Sources,
    # Sinks,
    Msource,
    Vsource,
    Vsink,
    SourceSink,
)
from pyschism.forcing.source_sink.nwm import NationalWaterModel

NWM = NationalWaterModel

__all__ = [
    # "Hydrology",
    "NationalWaterModel",
    # "Sources",
    # "Sinks",
    "Msource",
    "Vsource",
    "Vsink",
    "SourceSink",
    'NWM',
]



import argparse
import logging

from shapely.geometry import MultiPolygon

from pyschism.cmd import common
from pyschism.mesh.prop import Tvdflag, reg2multipoly


logger = logging.getLogger(__name__)


class TvdflagCli:

    def __init__(self, args: argparse.Namespace):

        # these are mutually exclusive:

        if args.default is not None:
            tvdflag = Tvdflag.default(args.hgrid)

        elif args.regfiles is not None:
            polygon_collection = []
            for regfile in args.regfiles:
                logger.info(f'Reading region file {regfile}...')
                for polygon in reg2multipoly(regfile).geoms:
                    polygon_collection.append(polygon)
            logger.info('Generating tvdflag from geometry...')
            tvdflag = Tvdflag.from_geometry(
                args.hgrid,
                MultiPolygon(polygon_collection)
            )

        elif args.geometry is not None:
            raise NotImplementedError('parse geometry')
            tvdflag = Tvdflag.from_geometry(
                args.hgrid,
                MultiPolygon(polygon_collection)
            )

        elif args.propfile is not None:
            element_values = []
            with open(args.propfile) as f:
                for i in range(len(args.hgrid.elements)):
                    element_values.append(int(f.readline().split()[-1]))
            tvdflag = Tvdflag(args.hgrid, element_values)

        if args.plot is True:
            tvdflag.make_plot(show=True)

        if args.output_path is not None:
            tvdflag.write(args.output_path, overwrite=args.overwrite)

    @staticmethod
    def add_subparser_action(subparsers):
        add_prop_options_to_parser(subparsers.add_parser('tvdflag'))


def add_prop_options_to_parser(parser):
    common.add_hgrid_to_parser(parser)
    common.add_log_level_to_parser(parser)
    parser.add_argument('--output-path', '-o')
    parser.add_argument('--overwrite', action='store_true')
    parser.add_argument('--plot', action='store_true')
    options = parser.add_mutually_exclusive_group(required=True)
    options.add_argument(
        '--default',
        '-d',
        help='Generate default tvdflag.prop file.'
    )
    options.add_argument(
        '--regfiles',
        '--reg',
        '-r',
        nargs='+',
        help='Uses ACE/gredit reg files.'
    )
    options.add_argument(
        '--propfile',
        '--prop',
        '-p',
        help='Loads a SCHISM prop file.'
    )
    options.add_argument(
        '--geometry',
        '--geom',
        '--gpd',
        '--json',
        '-g',
        nargs='+',
        help='Any file or files readable by geopandas. Must contain polygons '
             'or multipolygon geometries.'
    )



import argparse
from datetime import datetime, timedelta
import logging
import pathlib

from pyschism.cmd import common
from pyschism.forcing.bctides import Bctides


logger = logging.getLogger(__name__)


class BctidesCli:
    def __init__(self, args: argparse.Namespace):
        Bctides(
                args.hgrid,
                vgrid=args.vgrid,
                iettype=args.iettype,
                ifltype=args.ifltype,
                isatype=args.isatype,
                itetype=args.itetype,
                # itrtype=args.itrtype,
                cutoff_depth=args.cutoff_depth,
            ).write(
                args.output_directory,
                args.start_date,
                args.end_date,
                overwrite=args.overwrite,
                parallel_download=args.parallel_download
        )

    @staticmethod
    def add_subparser_action(subparsers):
        add_bctides_options_to_parser(subparsers.add_parser('bctides'))


def add_bctides_options_to_parser(parser):
    common.add_hgrid_to_parser(parser)
    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Allow overwrite of output files."
    )
    common.add_log_level_to_parser(parser)
    common.add_dates_to_parser(parser)
    # parser.add_argument(
    #     "--run-days",
    #     type=lambda x: timedelta(days=float(x)),
    #     required=True,
    #     help='Number of simulation days.'

    # )
    # parser.add_argument(
    #     "--start-date",
    #     type=lambda x: datetime.strptime(x, "%Y-%m-%dT%H:%M:%S"),
    #     # help=''
    #     # action=DateAction,
    # )
    # bctides.add_argument(
    #     "--date-formatting",
    #     type=str,
    #     type=lambda x: datetime.strptime(x, "%Y-%m-%dT%H:%M:%S"),
    #     # help=''
    # )
    parser.add_argument(
        "--output-directory",
        "-o",
        type=pathlib.Path,
        required=True
    )
    common.add_vgrid_to_parser(parser)
    common.add_tidal_constituents_to_parser(parser)
    common.add_tidal_database_to_parser(parser)
    common.add_baroclinic_database_to_parser(parser)
    common.add_bctides_options_to_parser(parser)
    common.add_ibctype_to_parser(parser)
    parser.add_argument("--parallel-download", action='store_true')


#     add_nudge_to_parser(parser)


# class NudgeAction(argparse.Action):
#     def __call__(self, parser, namespace, values, option_string=None):
        
#         if len(values) == 0:
#             values = None

#         if values is None:
#             if namespace.vgrid.is2D():
#                 values = False
#             else:
#                 values = True
#             if 'disable' in option_string:
#                 values = False

#         setattr(namespace, self.dest, values)


# def add_nudge_to_parser(parser):
#     parser.add_argument(
#         '--nudge-temp',
#         '--nudge-temperature',
#         '--nudge-t',
#         dest='nudge_temp',
#         nargs='*',
#         action=NudgeAction,
#     )
#     parser.add_argument(
#         '--disable-nudge-temp',
#         '--disable-nudge-temperature',
#         '--disable-nudge-t',
#         dest='nudge_temp',
#         nargs=0,
#         action=NudgeAction,
#     )
#     parser.set_defaults(nudge_temp=None)
#     parser.add_argument('--rlmax-temp', default=1.5, type=float)
#     parser.add_argument('--rnu_day-temp', default=0.25, type=float)
#     parser.add_argument(
#         '--nudge-salt',
#         '--nudge-salinity',
#         '--nudge-s',
#         dest='nudge_salt',
#         nargs='?',
#         action=NudgeAction,
#     )
#     parser.add_argument(
#         '--disable-nudge-salt',
#         '--disable-nudge-salinity',
#         '--disable-nudge-s',
#         dest='nudge_salt',
#         nargs='?',
#         action=NudgeAction,
#     )
#     parser.set_defaults(nudge_salt=None)
#     parser.add_argument('--rlmax-salt', default=1.5, type=float)
#     parser.add_argument('--rnu_day-salt', default=0.25, type=float)


# def get_tides(args: argparse.Namespace):
#     def get_elevation():
#         return True if args.tidal_database is not None else False

#     def get_velocity():
#         return (
#             True
#             if args.include_velocity is True or args.vgrid.is3D() is True
#             else False
#         )

#     return Tides(
#         elevation=get_elevation(),
#         velocity=get_velocity(),
#         tidal_database=args.tidal_database,
#     )


# class CustomBoundaryAction(argparse.Action):
#     def __call__(self, parser, namespace, values, option_string=None):
#         if isinstance(values, str):
#             if "iettype" in option_string:
#                 values = json.loads(values)
#                 for bnd_id, tval in values.items():
#                     if not isinstance(tval, int):
#                         raise TypeError(
#                             'Boundary type must be an integer, not type '
#                             f'{type(tval)}.')
#                     if tval == 3:
#                         namespace.hgrid.boundaries.set_forcing(
#                             bnd_id,
#                             iettype=iettype.Iettype3(get_tides(namespace))
#                             )
#                     else:
#                         raise ValueError
#         setattr(namespace, self.dest, values)


# class Iettype3Action(argparse.Action):

#     def __call__(self, parser, namespace, values, option_string=None):
#         for boundary in namespace.hgrid.boundaries.open.itertuples():
#             namespace.hgrid.boundaries.set_forcing(
#                 boundary.id,
#                 iettype=self.const(get_tides(namespace)))
#         setattr(namespace, self.dest, self.const)


# class Iettype4Action(argparse.Action):

#     def __call__(self, parser, namespace, values, option_string=None):
#         for boundary in namespace.hgrid.boundaries.open.itertuples():
#             namespace.hgrid.boundaries.set_forcing(
#                 boundary.id,
#                 iettype=self.const(baroclinic_databases[
#                     namespace.baroclinic_database]()))
#         setattr(namespace, self.dest, self.const)


# class Iettype5Action(argparse.Action):

#     def __call__(self, parser, namespace, values, option_string=None):
#         for boundary in namespace.hgrid.boundaries.open.itertuples():
#             namespace.hgrid.boundaries.set_forcing(
#                 boundary.id,
#                 iettype=self.const(
#                     get_tides(namespace),
#                     baroclinic_databases[namespace.baroclinic_database]()
#                     )
#             )
#         setattr(namespace, self.dest, self.const)


# class Ifltype3Action(argparse.Action):

#     def __call__(self, parser, namespace, values, option_string=None):
#         for boundary in namespace.hgrid.boundaries.open.itertuples():
#             namespace.hgrid.boundaries.set_forcing(
#                 boundary.id,
#                 ifltype=values(get_tides(namespace)))
#         setattr(namespace, self.dest, values)


# class Ifltype4Action(argparse.Action):

#     def __call__(self, parser, namespace, values, option_string=None):
#         for boundary in namespace.hgrid.boundaries.open.itertuples():
#             namespace.hgrid.boundaries.set_forcing(
#                 boundary.id,
#                 ifltype=self.const(baroclinic_databases[
#                     namespace.baroclinic_database]()))
#         setattr(namespace, self.dest, values)


# class Ifltype5Action(argparse.Action):
#     def __call__(self, parser, namespace, values, option_string=None):
#         for boundary in namespace.hgrid.boundaries.open.itertuples():
#             namespace.hgrid.boundaries.set_forcing(
#                 boundary.id,
#                 ifltype=self.const(
#                     get_tides(namespace),
#                     baroclinic_databases[namespace.baroclinic_database]()
#                     )
#             )
#         setattr(namespace, self.dest, values)


# class Itetype4Action(argparse.Action):
#     def __call__(self, parser, namespace, values, option_string=None):
#         tmp_parser = argparse.ArgumentParser(add_help=False)
#         add_nudge_to_parser(tmp_parser)
#         tmp_args, _ = tmp_parser.parse_known_args()
#         bnd_ids = namespace.hgrid.boundaries.open['id'].tolist()
#         if tmp_args.nudge_temp is None:
#             if namespace.vgrid.is2D():
#                 tmp_args.nudge_temp = []
#             else:
#                 tmp_args.nudge_temp = bnd_ids
#         elif tmp_args.nudge_temp is True:
#             tmp_args.nudge_temp = bnd_ids
#         elif tmp_args.nudge_temp is False:
#             tmp_args.nudge_temp = []
#         remaining_bounds = list(set(tmp_args.nudge_temp) - set(bnd_ids))
#         if len(remaining_bounds) > 0:
#             raise ValueError(f"No boundary with id's {remaining_bounds}.")
#         for boundary in namespace.hgrid.boundaries.open.itertuples():
#             namespace.hgrid.boundaries.set_forcing(
#                 boundary.id,
#                 itetype=self.const(
#                     baroclinic_databases[namespace.baroclinic_database](),
#                     ))
#             if boundary.id in tmp_args.nudge_temp:
#                 namespace.hgrid.boundaries.open.loc[boundary.Index].itetype.nudge = True
#                 namespace.hgrid.boundaries.open.loc[boundary.Index].itetype.rlmax = tmp_args.rlmax_temp               
#                 namespace.hgrid.boundaries.open.loc[boundary.Index].itetype.rnu_day = tmp_args.rnu_day_temp
#             else:
#                 namespace.hgrid.boundaries.open.loc[boundary.Index].itetype.nudge = False

#         setattr(namespace, self.dest, self.const)


# class Isatype4Action(argparse.Action):
#     def __call__(self, parser, namespace, values, option_string=None):
#         # logger.debug('Init isatype-4')
#         tmp_parser = argparse.ArgumentParser(add_help=False)
#         add_nudge_to_parser(tmp_parser)
#         tmp_args, _ = tmp_parser.parse_known_args()
#         bnd_ids = namespace.hgrid.boundaries.open['id'].tolist()
#         if tmp_args.nudge_salt is None:
#             if namespace.vgrid.is2D():
#                 tmp_args.nudge_salt = []
#             else:
#                 tmp_args.nudge_salt = bnd_ids
#         elif tmp_args.nudge_salt is True:
#             tmp_args.nudge_salt = bnd_ids
#         elif tmp_args.nudge_salt is False:
#             tmp_args.nudge_salt = []
#         # logger.debug('Init isatype-4')
#         remaining_bounds = list(set(tmp_args.nudge_salt) - set(bnd_ids))
#         if len(remaining_bounds) > 0:
#             raise ValueError(f"No boundary with id's {remaining_bounds}.")
#         for boundary in namespace.hgrid.boundaries.open.itertuples():
#             namespace.hgrid.boundaries.set_forcing(
#                 boundary.id,
#                 isatype=self.const(baroclinic_databases[
#                     namespace.baroclinic_database]()))
#             if boundary.id in tmp_args.nudge_salt:
#                 namespace.hgrid.boundaries.open.loc[boundary.Index].isatype.nudge = True
#                 namespace.hgrid.boundaries.open.loc[boundary.Index].isatype.rlmax = tmp_args.rlmax_salt               
#                 namespace.hgrid.boundaries.open.loc[boundary.Index].isatype.rnu_day = tmp_args.rnu_day_salt
#             else:
#                 namespace.hgrid.boundaries.open.loc[boundary.Index].isatype.nudge = False
#         setattr(namespace, self.dest, self.const)



from datetime import datetime, timedelta
from time import time
import logging

from pyschism.mesh.hgrid import Hgrid
from pyschism.forcing.nws.nws2.hrrr3 import HRRR
from pyschism.dates import nearest_cycle

logging.basicConfig(
    format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
    force=True,
)
logging.captureWarnings(True)

log_level = logging.DEBUG
logging.getLogger('pyschism').setLevel(log_level)

if __name__ == "__main__":
    t0 = time()
    rnday = 1
    record = 1
    startdate = nearest_cycle()
    hgrid = Hgrid.open('../../static/hgrid.gr3', crs='epsg:4326')

    pscr='/sciclone/pscr/lcui01/HRRR/'
    hrrr = HRRR(start_date=startdate, rnday=rnday, pscr=pscr, record=record, bbox=hgrid.bbox)
    print(f'It took {(time()-t0)/60} mins to process {rnday} days, {record*24} files/day')



from datetime import timedelta
import pathlib

# import geopandas as gpd
from netCDF4 import Dataset


class Elev2D:

    def __init__(self, bctides):
        self.bctides = bctides

    def write(
            self,
            elev2D,
            start_date,
            rnday,
            output_interval=timedelta(days=1),
            overwrite: bool = False,
            progress_bar: bool = True,
    ):
        elev2D = pathlib.Path(elev2D)
        if elev2D.exists() and overwrite is not True:
            raise IOError(f'File {elev2D} exists and overwrite is not True.')

        # file_is_not_needed = True
        timevec = None
        for boundary in self.bctides.gdf.itertuples():
            if boundary.iettype is not None:
                if boundary.iettype.iettype in [4, 5]:
                    ds = boundary.iettype.data_component
                    datasets = ds.get_datasets(
                                start_date,
                                rnday,
                                output_interval
                                )
                    timevec = range(len(datasets))

        if timevec is None:
            return

        nOpenBndNodes = 0
        for boundary in self.bctides.gdf.itertuples():
            nOpenBndNodes += len(boundary.indexes)

        dst = Dataset(elev2D, 'w', format='NETCDF4')
        # dimensions
        dst.createDimension('nOpenBndNodes', nOpenBndNodes)
        dst.createDimension('one', 1)
        dst.createDimension('time', None)
        dst.createDimension('nLevels', 1)
        dst.createDimension('nComponents', 1)

        # variables
        dst.createVariable('time_step', 'f', ('one',))
        dst['time_step'][:] = int(output_interval.total_seconds())
        dst.createVariable('time', 'f', ('time',))
        dst['time'][:] = timevec
        dst.createVariable('time_series', 'f',
                           ('time', 'nOpenBndNodes', 'nLevels', 'nComponents'))
        offset = 0
        for boundary in self.bctides.gdf.itertuples():
            if boundary.iettype is not None:
                if boundary.iettype.iettype in [4, 5]:
                    boundary.iettype.data_component.put_boundary_ncdata(
                        boundary, dst, start_date, rnday, overwrite=overwrite,
                        offset=offset, output_interval=output_interval,
                        pixel_buffer=10, progress_bar=progress_bar)
            else:
                self.put_null_boundary_data(dst, len(boundary.indexes))
            offset += len(boundary.indexes)

    def put_null_boundary_data(self, dst, np):
        raise NotImplementedError('Must write null data.')



from datetime import timedelta
import pathlib

# import geopandas as gpd
from netCDF4 import Dataset


class UV3D:

    def __init__(self, bctides):
        self.bctides = bctides

    def write(
            self,
            uv3d,
            start_date,
            rnday,
            output_interval=timedelta(days=1),
            overwrite: bool = False,
            progress_bar: bool = True,
    ):
        uv3d = pathlib.Path(uv3d)
        if uv3d.exists() and overwrite is not True:
            raise IOError(f'File {uv3d} exists and overwrite is not True.')

        # file_is_not_needed = True
        timevec = None
        for boundary in self.bctides.gdf.itertuples():
            if boundary.ifltype is not None:
                if boundary.ifltype.ifltype in [4, 5]:
                    ds = boundary.ifltype.data_component
                    datasets = ds.get_datasets(
                                start_date,
                                rnday,
                                output_interval
                                )
                    timevec = range(len(datasets))

        if timevec is None:
            return

        nOpenBndNodes = 0
        for boundary in self.bctides.gdf.itertuples():
            nOpenBndNodes += len(boundary.indexes)

        dst = Dataset(uv3d, 'w', format='NETCDF4')
        # dimensions
        dst.createDimension('nOpenBndNodes', nOpenBndNodes)
        dst.createDimension('one', 1)
        dst.createDimension('time', None)
        dst.createDimension('nLevels', self.bctides.vgrid.nvrt)
        dst.createDimension('nComponents', 2)

        # variables
        dst.createVariable('time_step', 'f', ('one',))
        dst['time_step'][:] = int(output_interval.total_seconds())
        dst.createVariable('time', 'f', ('time',))
        dst['time'][:] = timevec
        dst.createVariable('time_series', 'f',
                           ('time', 'nOpenBndNodes', 'nLevels', 'nComponents'))
        offset = 0
        for boundary in self.bctides.gdf.itertuples():
            if boundary.ifltype is not None:
                if boundary.ifltype.ifltype in [4, 5]:
                    boundary.ifltype.data_component.put_boundary_ncdata(
                        self.bctides.hgrid,
                        self.bctides.vgrid,
                        boundary,
                        dst,
                        start_date,
                        rnday,
                        overwrite=overwrite,
                        offset=offset,
                        output_interval=output_interval,
                        pixel_buffer=10,
                        progress_bar=progress_bar
                    )
            else:
                self.put_null_boundary_data(dst, len(boundary.indexes))
            offset += len(boundary.indexes)

    def put_null_boundary_data(self, dst, np):
        raise NotImplementedError('Must write null data.')



from abc import ABC, abstractmethod
from datetime import datetime, timedelta
import os
import pathlib
import shutil
import subprocess
from typing import Union

import f90nml

from pyschism import dates


class Hotstart(ABC):

    @property
    @abstractmethod
    def time(self) -> datetime:
        """Datetime object for the hotstart snapshot"""

    @property
    @abstractmethod
    def iteration(self) -> int:
        """Iteration number for hotstart snapshot"""

    @property
    @abstractmethod
    def path(self) -> pathlib.Path:
        """Path for output file."""

    @property
    def status(self):
        return self.path.exists()

    @staticmethod
    def combine(path):
        return CombineHotstartBinary(path)

    def move(self, dest, overwrite=False):
        self._path = shutil.move(self.path, dest)

    def symlink(self, dest, overwrite=False):
        hotstart_lnk = dest / 'hotstart.nc'
        if overwrite is True:
            if hotstart_lnk.exists():
                hotstart_lnk.unlink()
        os.symlink(os.path.relpath(
            self.hotstart.path, dest), hotstart_lnk)


class CombineHotstartBinary(Hotstart):

    def __init__(self, path: Union[str, os.PathLike], iteration=None,
                 binary='combine_hotstart7'):
        path = pathlib.Path(path)
        self.binary = binary
        if iteration is not None:
            self.iteration = iteration
        else:
            param = path / 'param.out.nml'
            if param.exists():
                param = f90nml.read(param)
            else:
                param = f90nml.read(path.parent / 'param.nml')
            self.iteration = int(
                timedelta(days=param['core']['rnday']) /
                timedelta(seconds=param['core']['dt']))
        self._path = pathlib.Path(path) / f"hotstart_it={self.iteration}.nc"

    def run(self):
        subprocess.check_call([
            f'{self.binary}', '-i', f'{self.iteration}'], cwd=self.path.parent)

    @property
    def time(self):
        param = f90nml.read(self.path.parent.parent / 'param.nml')
        start_hour = float(param['opt']['start_hour'])
        hours = int(start_hour)
        minutes = int((start_hour*60) % 60)
        seconds = int((start_hour*3600) % 60)
        return dates.localize_datetime(
            datetime(
                int(param['opt']['start_year']),
                int(param['opt']['start_month']),
                int(param['opt']['start_day']),
                hours, minutes, seconds
            )
            + timedelta(days=param['core']['rnday'])
            - timedelta(days=param['opt']['utc_start']))

    @property
    def iteration(self):
        return self._iteration

    @iteration.setter
    def iteration(self, iteration: int):
        self._iteration = int(iteration)

    @property
    def path(self):
        return self._path

    @property
    def binary(self) -> pathlib.Path:
        return self._binary

    @binary.setter
    def binary(self, binary):
        self._binary = pathlib.Path(binary)



import os
import copy

import numpy as np
from scipy import interpolate
import pandas as pd
from netCDF4 import Dataset

from pyschism.forcing.source_sink.sflux2source.TimeHistory import TimeHistory

def BinA(A, B):
    x = A
    y = B

    index = np.argsort(x)
    sorted_x = x[index]
    sorted_index = np.searchsorted(sorted_x, y)

    yindex = np.take(index, sorted_index, mode="clip")
    mask = x[yindex] != y

    result = np.ma.array(yindex, mask=mask)

    return [np.logical_not(np.ma.getmask(result)), np.ma.getmask(result)]


class SourceSinkIn():
    """ class for *.prop or other similar formats"""
    def __init__(self, filename=None, number_of_groups=2, ele_groups=[]):
        self.n_group = number_of_groups  # 0: source; 1: sink
        if filename is not None:
            """Read a usgs data file and initialize from its content """
            self.source_file = filename
            self.np_group = []
            self.ip_group = []

            with open(self.source_file) as fin:
                for k in range(0, self.n_group):
                    self.np_group.append(int(fin.readline().split()[0]))
                    print("Points in Group " + str(k+1) + ": " + str(self.np_group[k]))
                    self.ip_group.append(np.empty((self.np_group[k]), dtype=int))
                    for i in range(0, self.np_group[k]):
                        self.ip_group[k][i] = int(fin.readline())
                    fin.readline()  # empty line between groups
                    if self.np_group[k] > 0:
                        print("p first: " + str(self.ip_group[k][0]))
                        print("p last: " + str(self.ip_group[k][-1]))
        else:
            self.np_group = [len(x) for x in ele_groups]
            self.ip_group = [np.array(x) for x in ele_groups]

    def writer(self, filename=None):
        if filename is None:
            filename = self.source_file

        with open(filename, 'w') as fout:
            for k in range(0, self.n_group):
                print("Points in Group " + str(k+1) + ": " + str(self.np_group[k]))
                fout.write(f"{self.np_group[k]}\n")
                for i in range(0, self.np_group[k]):
                    fout.write(f"{self.ip_group[k][i]}\n")
                fout.write("\n")  # empty line

    def add_points(self, added_points=np.array([]), group=None):
        self.np_group[group] += len(added_points)
        self.ip_group[group] = np.r_[self.ip_group[group], added_points]

    def get_max(self):
        """get max prop value"""

    def get_ele(self, pval):
        """return ele id with specified prop value"""


class source_sink():
    """class for handling all source/sink inputs"""
    def __init__(self, source_dir=None, start_time_str='2000-01-01 00:00:00', timedeltas=[0.0, 86400.0*365],
                 source_eles=[], sink_eles=[], vsource_data=None, vsink_data=None):

        dummy_source_sink = TimeHistory(
            file_name=None, start_time_str='2000-01-01 00:00:00',
            data_array=np.c_[np.array(timedeltas), np.array(timedeltas)*0.0], columns=['datetime', '1']
        )

        if source_dir is None:
            nt = len(timedeltas)
            nsources = len(source_eles)
            nsinks = len(sink_eles)

            if vsource_data is None:
                vsource_data = np.zeros([nt, nsources])
            if vsink_data is None:
                vsink_data = np.zeros([nt, nsinks])

            if nsources > 0:
                self.vsource = TimeHistory(file_name=None, start_time_str=start_time_str,
                                           data_array=np.c_[np.array(timedeltas), vsource_data],
                                           columns=['datetime']+source_eles)
                self.msource = TimeHistory(file_name=None, start_time_str=start_time_str,
                                           data_array=np.c_[np.array(timedeltas), -9999*np.ones([nt, nsources]), np.zeros([nt, nsources])],
                                           columns=['datetime']+source_eles+source_eles)
            else:
                source_eles = [1]  # dummy with 0 vsource
                self.vsource = copy.deepcopy(dummy_source_sink)
                self.msource = TimeHistory(file_name=None, start_time_str=start_time_str,
                                           data_array=np.c_[np.array(timedeltas), -9999*np.ones([nt, 1]), np.zeros([nt, 1])],
                                           columns=['datetime', '1', '1'])

            if nsinks > 0:
                self.vsink = TimeHistory(file_name=None, start_time_str=start_time_str,
                                         data_array=np.c_[np.array(timedeltas), vsink_data],
                                         columns=['datetime']+sink_eles)
            else:
                sink_eles = [1]  # dummy with 0 vsource
                self.vsink = copy.deepcopy(dummy_source_sink)

            self.source_sink_in = SourceSinkIn(filename=None, number_of_groups=2, ele_groups=[source_eles, sink_eles])

        else:  # read from existing files
            self.source_sink_in = SourceSinkIn(f"{source_dir}/source_sink.in")
            self.source_dir = source_dir

            if self.source_sink_in.np_group[0] > 0:
                print('reading vsource\n')
                self.vsource = TimeHistory(f"{source_dir}/vsource.th", start_time_str=start_time_str, columns=['datetime']+self.source_sink_in.ip_group[0].tolist())
                print('reading msource\n')
                self.msource = TimeHistory(f"{source_dir}/msource.th", start_time_str=start_time_str, columns=['datetime']+self.source_sink_in.ip_group[0].tolist()+self.source_sink_in.ip_group[0].tolist())
                source_eles = self.source_sink_in.ip_group[0]
            else:
                self.vsource = copy.deepcopy(dummy_source_sink)
                self.msource = TimeHistory(file_name=None, start_time_str=start_time_str,
                                           data_array=np.c_[np.array(timedeltas), -9999*np.ones([nt, 1]), np.zeros([nt, 1])],
                                           columns=['datetime', '1', '1'])
                source_eles = [1]


            if self.source_sink_in.np_group[1] > 0:
                print('reading vsink\n')
                self.vsink = TimeHistory(f"{source_dir}/vsink.th", start_time_str=start_time_str, columns=['datetime']+self.source_sink_in.ip_group[1].tolist())
                self.vsink.df.columns = self.vsink.df.columns.map(str)
                sink_eles = self.source_sink_in.ip_group[1]
            else:  # dummy vsink
                self.vsink = copy.deepcopy(dummy_source_sink)
                sink_eles = [1]

            # accomodate dummy sources/sinks
            self.source_sink_in = SourceSinkIn(filename=None, number_of_groups=2, ele_groups=[source_eles, sink_eles])

        self.vsource.df.columns = self.vsource.df.columns.map(str)
        self.msource.df.columns = self.msource.df.columns.map(str)
        self.vsink.df.columns = self.vsink.df.columns.map(str)

        self.nsource = self.source_sink_in.np_group[0]
        self.nsink = self.source_sink_in.np_group[1]
        self.source_eles = source_eles
        self.sink_eles = sink_eles
        self.ntracers = 2
    
    def update_vars(self):
        self.vsource.df_propagate()
        self.msource.df_propagate()
        self.vsink.df_propagate()
        self.nsource = self.source_sink_in.np_group[0]
        self.nsink = self.source_sink_in.np_group[1]
        self.source_eles = self.source_sink_in.ip_group[0]
        self.sink_eles = self.source_sink_in.ip_group[1]
        self.ntracers = 2
        
        self.vsource.data = self.vsource.df.values[:, 1:]

    def __add__(self, other):
        '''
        Add source_sink B to source_sink A,
        retaining A's time stamps
        '''
        A = copy.deepcopy(self)
        B = other
        for i, source_sink in enumerate(['vsource', 'vsink']):  # 0: source; 1: sink
            print(f'Adding type {i}:')
            #if vsink only contains dummy sink, then skip interpolation (todo: remove dummy sinks)
            if i == 1 and len(A.source_sink_in.ip_group[i]) <= 1: continue
            [_, new_eles_inds] = BinA(A.source_sink_in.ip_group[i],
                                      B.source_sink_in.ip_group[i])
            [existing_eles_inds, _] = BinA(B.source_sink_in.ip_group[i],
                                           A.source_sink_in.ip_group[i])
            new_eles = B.source_sink_in.ip_group[i][new_eles_inds]
            existing_eles = A.source_sink_in.ip_group[i][existing_eles_inds]

            A.source_sink_in.add_points(added_points=new_eles, group=i)

            B_ss = eval(f"B.{source_sink}")
            A_ss = eval(f"A.{source_sink}")

            #use A's time orgin
            time_diff = (B_ss.df.iloc[0, 0] - A_ss.df.iloc[0, 0]).total_seconds()

            f_interp = interpolate.interp1d(B_ss.time + time_diff, B_ss.df.iloc[:, 1:].to_numpy().T)
            B_df_interp = pd.DataFrame(data=f_interp(A_ss.time).T, columns=B_ss.df.columns[1:])

            # clip the values in case of small negative/positive values due to trunction error during interpolation
 
            if i == 0:  # source
                B_df_interp = B_df_interp.clip(0.0, None)
            else:
                B_df_interp = B_df_interp.clip(None, 0.0)
 
            A_ss.df = copy.deepcopy(A_ss.df.join(B_df_interp[new_eles.astype(str)]))
            A_ss.df[existing_eles.astype('str')] += B_df_interp[existing_eles.astype('str')]

            A_ss.n_time = A_ss.df.shape[0]
            A_ss.n_station = A_ss.df.shape[1] - 1

        A.msource = TimeHistory(
            file_name=None,
            data_array=np.c_[np.array([0.0, 86400*365*100]), -9999*np.ones([2, len(A.vsource.df.columns)-1]), np.zeros([2, len(A.vsource.df.columns)-1])],
            columns=['datetime']+A.vsource.df.columns[1:].tolist()+A.vsource.df.columns[1:].tolist())

        A.update_vars()

        return A

    def writer(self, dirname=None):
        if dirname is None:
            raise Exception("dirname is required.")
        os.makedirs(dirname, exist_ok=True)

    def nc_writer(self, dirname=None):
        if dirname is None:
            raise Exception("dirname is required.")
        os.makedirs(dirname, exist_ok=True)

        msource_data = self.msource.data.reshape([self.msource.n_time, self.ntracers, self.nsource])
        fout = Dataset(f'{dirname}/source.nc', 'w', format='NETCDF4')
        fout.createDimension('nsources', self.nsource)
        fout.createDimension('nsinks', self.nsink)
        fout.createDimension('ntracers', self.ntracers)
        fout.createDimension('time_msource', self.msource.n_time)
        fout.createDimension('time_vsource', self.vsource.n_time)
        fout.createDimension('time_vsink', self.vsink.n_time)
        fout.createDimension('one', np.atleast_1d(1))

        fout.createVariable('source_elem', 'i', ('nsources'))
        fout['source_elem'][:] = self.source_sink_in.ip_group[0]

        fout.createVariable('vsource', 'f', ('time_vsource', 'nsources'))
        fout['vsource'][:,:] = self.vsource.data

        fout.createVariable('msource', 'f', ('time_msource', 'ntracers', 'nsources'))
        fout['msource'][:,:,:] = msource_data

        fout.createVariable('sink_elem', 'i', ('nsinks'))
        fout['sink_elem'][:] = self.source_sink_in.ip_group[1]

        fout.createVariable('vsink', 'f', ('time_vsink', 'nsinks'))
        fout['vsink'][:,:] = self.vsink.data
        #C.vars.extend(['time_step_vsource','time_step_msource','time_step_vsink'])
   
        fout.createVariable('time_step_vsource', 'f', ('one'))
        fout['time_step_vsource'][:] = self.vsource.delta_t

        fout.createVariable('time_step_msource', 'f', ('one'))
        fout['time_step_msource'][:] = self.msource.delta_t

        fout.createVariable('time_step_vsink', 'f', ('one'))
        fout['time_step_vsink'][:] = self.vsink.delta_t

        fout.close()



from datetime import datetime, timedelta
from functools import cached_property, lru_cache
import pathlib
from typing import Dict, Union
import logging

from ordered_set import OrderedSet
import numpy as np

from pyschism import dates
from pyschism.mesh.vgrid import Vgrid
from pyschism.forcing.bctides import iettype, ifltype, isatype, itetype, itrtype, Tides
from pyschism.forcing.bctides.elev2d import Elev2D
from pyschism.forcing.bctides.uv3d import UV3D
from pyschism.forcing.bctides.mod3d import TEM_3D, SAL_3D

logger = logging.getLogger(__name__)


class IbctypeDescriptor:
    def __init__(self, name, bctype):
        self.name = name
        self.bctype = bctype

    def __get__(self, obj, val):
        return obj.gdf[self.name]

    def __set__(self, obj, val):
        if val is not None:
            if isinstance(val, dict):
                for bnd_id, ibctype in val.items():
                    if not isinstance(ibctype, (self.bctype, type(None))):
                        raise TypeError(
                            f"Argument {ibctype} for boundary {bnd_id} must be of type {self.bctype} "
                            f" or None, not type {type(ibctype)}."
                        )
                    obj.gdf.at[np.where(obj.gdf["id"] == bnd_id)[0][0], self.bctype.__name__.lower()] = ibctype
            else:
                if not isinstance(val, self.bctype):
                    raise TypeError(
                        f"Argument {self.name} must be of type "
                        f"{self.bctype}, not type {type(val)}."
                    )
                obj.gdf[self.name] = val


class BctidesMeta(type):
    def __new__(meta, name, bases, attrs):
        bctypes = {
            "iettype": iettype.Iettype,
            "ifltype": ifltype.Ifltype,
            "isatype": isatype.Isatype,
            "itetype": itetype.Itetype,
            "itrtype": itrtype.Itrtype,
        }
        for name, ibctype in bctypes.items():
            attrs[name] = IbctypeDescriptor(name, ibctype)
        return type(name, bases, attrs)


class Bctides(metaclass=BctidesMeta):

    start_date = dates.StartDate()
    end_date = dates.EndDate()
    run_days = dates.RunDays()

    def __init__(
        self,
        hgrid,
        vgrid=None,
        iettype: Union[Dict, iettype.Iettype] = None,
        ifltype: Union[Dict, ifltype.Ifltype] = None,
        isatype: Union[Dict, isatype.Isatype] = None,
        itetype: Union[Dict, itetype.Itetype] = None,
        itrtype: Union[Dict, itrtype.Itrtype] = None,
        cutoff_depth: float = 50.0,
    ):
        self.hgrid = hgrid
        self.vgrid = Vgrid.default() if vgrid is None else vgrid
        self.cutoff_depth = cutoff_depth
        self.iettype = iettype
        self.ifltype = ifltype
        self.isatype = isatype
        self.itetype = itetype
        self.itrtype = itrtype

    def __str__(self):
        f = [
            f"{str(self.start_date)}",
            f"{self.ntip} {self.cutoff_depth}",
        ]
        if self.ntip > 0:
            for constituent in self.tides.get_active_potential_constituents():
                forcing = self.tides(self.start_date, self.rnday, constituent)
                f.append(
                    " ".join(
                        [
                            f"{constituent}\n",
                            f"{forcing[0]:G}",
                            f"{forcing[1]:G}",
                            f"{forcing[2]:G}",
                            f"{forcing[3]:G}",
                            f"{forcing[4]:G}",
                        ]
                    )
                )
        f.append(f"{self.nbfr:d}")
        if self.nbfr > 0:
            for constituent in self.tides.get_active_forcing_constituents():
                forcing = self.tides(self.start_date, self.rnday, constituent)
                f.append(
                    " ".join(
                        [
                            f"{constituent}\n",
                            f"{forcing[2]:G}",
                            f"{forcing[3]:G}",
                            f"{forcing[4]:G}",
                        ]
                    )
                )
        global_constituents = self.tides.get_active_constituents()
        f.append(f"{len(self.gdf)}")
        for boundary in self.gdf.itertuples():
            f.append(self.get_forcing_string(boundary, global_constituents))
        return "\n".join(f)

    def write(
        self,
        output_directory,
        start_date: datetime = None,
        end_date: Union[datetime, timedelta] = None,
        bctides: Union[bool, str] = True,
        elev2D: Union[bool, str] = False,
        uv3D: Union[bool, str] = False,
        tem3D: Union[bool, str] = False,
        sal3D: Union[bool, str] = False,
        overwrite: bool = False,
        parallel_download=False,
        progress_bar=True,
    ):
        if start_date is not None:
            self.start_date = start_date
        if end_date is not None:
            self.run_days = end_date
        # self.tidal_database.write(path, )
        output_directory = pathlib.Path(output_directory)
        output_directory.mkdir(exist_ok=overwrite, parents=True)
        bctides = output_directory / "bctides.in" if bctides is True else bctides
        if bctides.exists() and not overwrite:
            raise IOError("path exists and overwrite is False")
        with open(bctides, "w") as f:
            f.write(str(self))
        # write nudge
        for bctype, tracer in {"itetype": "TEM", "isatype": "SAL"}.items():
            for boundary in self.gdf.itertuples():
                data_source = getattr(boundary, bctype)
                if data_source is not None:
                    import importlib
                    if hasattr(data_source, 'rlmax'):
                        # This generates all the nudges and writes the nudge files.
                        nudgemod = importlib.import_module('pyschism.forcing.bctides.nudge')
                        nudgeclass = getattr(nudgemod, f'{tracer}_Nudge')
                        _tracerfile = locals()[f'{tracer.lower()}3D']
                        if _tracerfile is False:
                            continue
                        elif _tracerfile is True:
                            _tracerfile = output_directory / f'{tracer}_nudge.gr3'
                        nudgeclass(self, data_source, rlmax=data_source.rlmax, rnu_day=data_source.rnu_day
                                   ).write(_tracerfile, overwrite=overwrite)
                        break

        def write_elev2D():
            _elev2D = output_directory / "elev2D.th.nc" if elev2D is True else elev2D
            Elev2D(self).write(
                _elev2D,
                self.start_date,
                self.rnday,
                timedelta(days=1),
                overwrite,
                progress_bar=progress_bar,
            )

        def write_uv3D():
            # write uv3D.th.nc
            _uv3D = output_directory / "uv3D.th.nc" if uv3D is True else uv3D
            UV3D(self).write(
                _uv3D,
                self.start_date,
                self.rnday,
                timedelta(days=1),
                overwrite,
                progress_bar=progress_bar,
            )

        def write_tem3D():
            # write TEM_3D.th.nc
            _tem3D = output_directory / "TEM_3D.th.nc" if tem3D is True else tem3D
            TEM_3D(self).write(
                _tem3D,
                self.start_date,
                self.rnday,
                timedelta(days=1),
                overwrite,
                progress_bar=progress_bar,
            )

        def write_sal3D():
            _sal3D = output_directory / "SAL_3D.th.nc" if sal3D is True else sal3D
            SAL_3D(self).write(
                _sal3D,
                self.start_date,
                self.rnday,
                timedelta(days=1),
                overwrite,
                progress_bar=progress_bar,
            )

        if parallel_download is True:
            from multiprocessing import Process

            jobs = [
                Process(target=fn)
                for fn, fl in (
                    (write_elev2D, elev2D),
                    (write_uv3D, uv3D),
                    (write_tem3D, tem3D),
                    (write_sal3D, sal3D)
                ) if fl
            ]
            for job in jobs:
                job.start()
            for job in jobs:
                job.join()
            if any(j.exitcode != 0 for j in jobs):
                raise RuntimeError("Some parallel writer jobs failed!")
        else:
            if elev2D:
                write_elev2D()
            if uv3D:
                write_uv3D()
            if tem3D:
                write_tem3D()
            if sal3D:
                write_sal3D()

        # def write_tracer(tracer):
        #     tracer.write()

        # for tracer in [self.temperature, self.salinity, *self.tracers]:
        #     if tracer is not None:
        #         write_tracer(tracer)

    def get_forcing_string(self, boundary, global_constituents):

        bctypes = [
            boundary.iettype,
            boundary.ifltype,
            boundary.itetype,
            boundary.isatype,
        ]

        def get_focing_digit(bctype):
            if bctype is not None:
                return bctype.forcing_digit
            return "0"

        line = [
            f"{len(boundary.indexes)}",
            *[str(digit) for digit in map(get_focing_digit, bctypes)],
        ]
        f = [" ".join(line)]
        for bctype in bctypes:
            if bctype is not None:
                f.append(
                    bctype.get_boundary_string(
                        self.hgrid, boundary, global_constituents=global_constituents
                    )
                )
        return "\n".join(f)

    @property
    def rnday(self):
        return self.run_days

    @cached_property
    def gdf(self):
        gdf = self.hgrid.boundaries.open.copy()
        gdf["iettype"] = None
        gdf["ifltype"] = None
        gdf["isatype"] = None
        gdf["itetype"] = None
        gdf["itrtype"] = None
        return gdf

    @property
    def ntip(self):
        return len(self.tides.get_active_potential_constituents())

    @property
    def nbfr(self):
        return len(self.tides.get_active_forcing_constituents())

    @cached_property
    def tides(self):
        return TidalConstituentCombiner(self.gdf)


class TidalConstituentCombiner(Tides):
    def __init__(self, gdf):
        self.gdf = gdf
        afc = self.get_active_forcing_constituents()
        apc = self.get_active_potential_constituents()
        for constituent in set([*afc, *apc]):
            self.use_constituent(
                constituent,
                forcing=True if constituent in afc else False,
                potential=True if constituent in apc else False,
            )

    def get_active_forcing_constituents(self):
        active_constituents = OrderedSet()
        for row in self.gdf.itertuples():
            if row.iettype is not None:
                if row.iettype.iettype in [3, 5]:
                    [
                        active_constituents.add(x)
                        for x in row.iettype.tides.get_active_constituents()
                    ]
            if row.ifltype is not None:
                if row.ifltype.ifltype in [3, 5]:
                    [
                        active_constituents.add(x)
                        for x in row.ifltype.tides.get_active_constituents()
                    ]

        return list(active_constituents)

    def get_active_potential_constituents(self):
        active_constituents = OrderedSet()
        for row in self.gdf.itertuples():
            if row.iettype is not None:
                if row.iettype.iettype in [3, 5]:
                    [
                        active_constituents.add(x)
                        for x in row.iettype.tides.get_active_potential_constituents()
                    ]
            if row.ifltype is not None:
                if row.ifltype.ifltype in [3, 5]:
                    [
                        active_constituents.add(x)
                        for x in row.ifltype.tides.get_active_potential_constituents()
                    ]

        return list(active_constituents)

    @lru_cache
    def get_active_constituents(self):
        return list(
            OrderedSet(
                [
                    *self.get_active_potential_constituents(),
                    *self.get_active_forcing_constituents(),
                ]
            )
        )


def ad_hoc_test():
    from datetime import datetime
    import logging

    from pyschism.mesh import Hgrid
    from pyschism.forcing.bctides import Bctides, iettype, ifltype, isatype, itetype

    # setup logging
    logging.basicConfig(
        format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
        force=True,
    )
    logging.getLogger("pyschism").setLevel(logging.DEBUG)

    startdate = datetime(2018, 8, 17)
    print(startdate)
    rnday = 61
    hgrid = Hgrid.open("./hgrid.gr3", crs="epsg:4326")

    # Bctides
    iet3 = iettype.Iettype3(constituents='major', database='tpxo')
    iet4 = iettype.Iettype4()
    iet5 = iettype.Iettype5(iettype3=iet3, iettype4=iet4)
    ifl3 = ifltype.Ifltype3(constituents='major', database='tpxo')
    ifl4 = ifltype.Ifltype4()
    ifl5 = ifltype.Ifltype5(ifltype3=ifl3, ifltype4=ifl4)
    isa3 = isatype.Isatype4()
    # ite3 = itetype.Itetype4()
    bctides = Bctides(hgrid, iettype={'1': iet5}, ifltype={'1': ifl5},
                      isatype=isa3,
                      itetype={'1': itetype.Itetype2(10, 1)}
                      )
    bctides.write(
            './',
            startdate,
            rnday,
            bctides=True,
            elev2D=False,
            uv3D=False,
            tem3D=False,
            sal3D=False,
            overwrite=True
            )


if __name__ == "__main__":
    ad_hoc_test()



from datetime import datetime, timedelta
from time import time
import multiprocessing as mp
import logging

import numpy as np
import pandas as pd

from pyschism.mesh.hgrid import Hgrid
from pyschism.forcing.nws.nws2.gfs2 import GFS
from pyschism.dates import nearest_cycle

logging.basicConfig(
    format="[%(asctime)s] %(name)s %(levelname)s: %(message)s",
    force=True,
)
logging.captureWarnings(True)

log_level = logging.DEBUG
logging.getLogger('pyschism').setLevel(log_level)

if __name__ == "__main__":

    t0 = time()
    startdate = datetime(2021, 10, 13)

    pscr = '/sciclone/pscr/lcui01/GFS/'
    rnday = 5
    record = 1

    hgrid = Hgrid.open('../../static/hgrid.gr3', crs='epsg:4326')

    gfs = GFS(start_date=startdate, rnday=rnday, pscr=pscr, record=record, bbox=hgrid.bbox)

    print(f'It took {(time()-t0)/60} mins to process {rnday} days, {record*24} records/day')



from abc import abstractmethod
from enum import Enum

from pyschism.forcing.bctides.bctypes import Bctype
from pyschism.forcing import hycom


class Isatype(Bctype):

    @property
    @abstractmethod
    def isatype(self) -> int:
        pass

    @property
    def forcing_digit(self):
        return self.isatype


class UniformTimeHistorySalinity(Isatype):
    def __init__(self, time_history):
        raise NotImplementedError(f"{self.__class__.__name__}")
        self.time_history = time_history

    @property
    def isatype(self) -> int:
        return 1


class ConstantSalinity(Isatype):
    def __init__(self, value: float, nudging_factor: float):
        self.value = value
        if not (nudging_factor >= 0) and (nudging_factor <= 1):
            raise ValueError(f'Argument `nudging_factor` must be get 0 and let 1, but got {nudging_factor}.')
        self.nudging_factor = nudging_factor
        super().__init__()


    @property
    def isatype(self) -> int:
        return 2

    def get_boundary_string(self, *args, **kwargs) -> str:
        boundary_string = [
            f'{self.value:0.6f}',
            f'{self.nudging_factor:0.6f}',
        ]
        return '\n'.join(boundary_string)


class SalinityInitialConditions(Isatype):
    def __init__(self):
        raise NotImplementedError(f"{self.__class__.__name__}")

    @property
    def isatype(self):
        return 3


class SpatiallyVaryingTimeHistorySalinity(Isatype):
    class BaroclinicDatabases(Enum):
        #RTOFS = hycom.RTOFS
        GOFS = hycom.GOFS

    def __init__(
        self,
        data_source="gofs",
        nudge: bool = True,
        rlmax=1.5,
        rnu_day=0.25,
    ):
        if isinstance(data_source, str):
            data_source = self.BaroclinicDatabases[data_source.upper()].value()
        if not isinstance(data_source, hycom.Hycom):
            raise TypeError(
                "Argument data_source must be of type str or type "
                f"{type(hycom.Hycom)}, not type {type(data_source)}."
            )
        self.data_source = data_source
        self.data_component = data_source.salinity
        self.nudge = nudge
        self.rlmax = rlmax
        self.rnu_day = rnu_day

    def get_boundary_string(self, *args, **kwargs):
        return "1."

    @property
    def isatype(self):
        return 4


class Isatype1(UniformTimeHistorySalinity):
    pass


class Isatype2(ConstantSalinity):
    pass


class Isatype3(SalinityInitialConditions):
    pass


class Isatype4(SpatiallyVaryingTimeHistorySalinity):
    pass



from datetime import datetime, timedelta
from typing import Union

import numpy as np
import pytz


class StartDate:

    def __init__(self):
        self.start_date = None

    def __set__(self, obj, start_date: datetime):
        if start_date is not None:
            start_date = localize_datetime(start_date).astimezone(pytz.utc)

            if obj.end_date is not None:
                if start_date > obj.end_date:
                    raise ValueError(
                        'start_date is greater than end_date. '
                        'Try clearing end_date first.')
        self.start_date = start_date

    def __get__(self, obj, val) -> datetime:
        return self.start_date

    def __delete__(self, obj):
        self.start_date = None


class RunDays:

    def __init__(self):
        self.run_days = None

    def __set__(self, obj, run_days: Union[timedelta, float]):

        if isinstance(run_days, (int, float)):
            run_days = timedelta(days=float(run_days))
        
        elif isinstance(run_days, datetime):
            run_days = localize_datetime(run_days) - obj.start_date
            
        elif not isinstance(run_days, timedelta):
            raise TypeError(
                f'Argument run_days must be {timedelta} or float, not {run_days}.')
        
        if obj.start_date is not None:
            obj.end_date = obj.start_date + run_days

        self.run_days = run_days

    def __get__(self, obj, val) -> Union[timedelta, None]:
        return self.run_days

    def __delete__(self, obj):
        self.run_days = None


class EndDate:

    def __init__(self):
        self.end_date = None

    def __set__(self, obj, val: Union[float, timedelta, datetime]):

        if isinstance(val, (int, float)):
            val = timedelta(days=float(val))

        if isinstance(val, datetime):
            val = localize_datetime(val).astimezone(pytz.utc)

        elif not isinstance(val, timedelta) and obj.start_date is not None:
            val = obj.start_date + timedelta(days=float(val))

        elif isinstance(val, timedelta) and obj.start_date is not None:
            val = obj.start_date + val

        else:
            val = None

        self.end_date = val

    def __get__(self, obj, val) -> datetime:
        return self.end_date

    def __delete__(self, obj):
        self.end_date = None


class SpinupTime:

    def __init__(self):
        self.spinup_time = None

    def __set__(self, obj, val: Union[int, float, timedelta]):
        if not isinstance(val, timedelta):
            val = timedelta(days=float(val))
        self.spinup_time = val

    def __get__(self, obj, val) -> timedelta:
        return self.spinup_time

    def __delete__(self, obj):
        self.spinup_time = None


def nearest_zulu(input_datetime=None, method='floor'):
    """
    "pivot time" is defined as the nearest floor t00z for any given datetime.
    If this function is called without arguments, it will return the pivot time
    for the current datetime in UTC.
    """
    input_datetime = nearest_cycle(method=method) if input_datetime is None \
        else localize_datetime(input_datetime).astimezone(pytz.utc)
    return localize_datetime(
        datetime(input_datetime.year, input_datetime.month, input_datetime.day)
    )


def nearest_cycle(input_datetime=None, period=6, method='floor'):
    # Bug: method='ceil' doesn´t work correctly for cycle 18
    if input_datetime is None:
        input_datetime = localize_datetime(datetime.utcnow())
    assert method in ['floor', 'ceil']
    if method == 'floor':
        method = np.floor
    if method == 'ceil':
        method = np.ceil
    if isinstance(period, timedelta):
        period /= timedelta(hours=1)

    current_cycle = int(period * method(input_datetime.hour / period)) % 24

    return pytz.timezone('UTC').localize(
        datetime(input_datetime.year, input_datetime.month,
                 input_datetime.day, current_cycle))


def localize_datetime(d):
    # datetime is naïve iff:
    if d.tzinfo is None or d.tzinfo.utcoffset(d) is None:
        return pytz.timezone('UTC').localize(d)
    return d


def utcnow():
    return localize_datetime(datetime.utcnow())


def round_time(dt: datetime, date_delta: timedelta):
    """Round a datetime object to a multiple of a timedelta
    d : datetime.datetime object, default now.
    date_delta : timedelta object, we round to a multiple of this, default 1 minute.
    Author: Thierry Husson 2012 - Use it as you want but don't blame me.
            Stijn Nevens 2014 - Changed to use only datetime objects as variables
    https://stackoverflow.com/questions/3463930/how-to-round-the-minute-of-a-datetime-object
    """
    d = dt.replace(tzinfo=None)
    roundTo = date_delta.total_seconds()
    seconds = (d - d.min).seconds
    # // is a floor division, not a comment on following line:
    rounding = (seconds+roundTo/2) // roundTo * roundTo
    return d + timedelta(0, rounding-seconds, - d.microsecond)
